// 
//
//////////////////////////////////////////////////////////////////////

#include <malloc.h>
#include "classic.h"
#include "public.h"

//////////////////////////////////////////////////////////////////////
// 
//////////////////////////////////////////////////////////////////////



//    ＝＝＝ 《测量平差程序设计》－第三章 经典平差 源程序 ＝＝＝
//    ＝＝＝ 《测量平差程序设计》－第三章 经典平差 源程序 ＝＝＝


//////////////////////////////////////////////////////////////////////
//  参数平差程序(观测值独立)
// 参数：
// int n - 观测值个数
// int t - 参数个数
// double A[] - 误差方程系数矩阵，数组长度为n×t
// double L[] - 观测值向量,数组长度为n
// double P[] - 观测值权矩阵的对角线元素,数组长度为n
// double X[] - 参数向量,函数调用前为参数的先验值,调用后为参数平差值，
//              数组长度为t
// double Qx[] - 函数调用前为参数的先验权逆阵，调用后为参数平差值的权逆阵，
//               仅存下三角矩阵，数组长度为t×(t+1)/2 
// double V[] - 残差向量，数组长度为n
// 返回值：验后单位权中误差
//////////////////////////////////////////////////////////////////////////
double Adjust(int n,int t, double A[],double L[],double P[],
			   double X[],double N[],double V[])
{
	double *U=(double *)malloc(sizeof(double)*t); //临时数组，存法方程自由项
	
	//  法方程系数矩阵与自由项清零 
	for(int i=0;i<t*(t+1)/2;i++) N[i]=0.0;
	for(int j=0;j<t;j++) U[j]=0.0;
	
	//  组成法方程式 
	for(int k=0;k<n;k++)
	{
		for(int i=0;i<t;i++)
		{
			U[i]+=A[k*t+i]*P[k]*L[k];
			for(int j=0; j<=i; j++)
			{
				N[ij(i,j)]+=A[k*t+i]*P[k]*A[k*t+j];
			}
		}
	}

	//  计 算 X 	
	if(!inverse(N,t))
	{
		free(U); 
		return -1.0;
	}
	for(int i=0;i<t;i++)
	{
		X[i]=0.0;
		for(int j=0;j<t;j++)X[i]+=N[ij(i,j)]*U[j];
	}
	free(U);
	
	//   计 算 V 
	double pvv=0;
	for(int i=0;i<n;i++)
	{
		double vi=-L[i];
		for(int j=0;j<t;j++) vi+=A[i*t+j]*X[j];
		V[i]=vi;
		pvv+=vi*vi*P[i];
	}
	
	if(n==t)return 0.0; //n=t,无多余观测
	return sqrt(pvv/(n-t));

}






//////////////////////////////////////////////////////////////////////////
//          相关观测值参数平差
// 参数：
// int n - 观测值个数
// int t - 参数个数
// double A[] - 误差方程系数矩阵，数组长度为n×t
// double L[] - 观测值向量,数组长度为n
// double P[] - 观测值权矩阵的对角线元素,数组长度为n
// double X[] - 参数向量的平差值,数组长度为t
// double N[] - 参数平差值的权逆阵，仅存下三角矩阵，数组长度为t×(t+1)/2 
// double V[] - 残差向量，数组长度为n
// 返回值：验后单位权中误差
//////////////////////////////////////////////////////////////////////////

double Co_Adjust(int n, int t,double A[],double L[],double P[],
			     double X[],double N[],double V[])
{
	double *U=(double *)malloc(sizeof(double)*t);
	
	//法方程系数阵和自由项清零
	for(int i=0; i<t*(t+1)/2; i++) N[i]=0.0;
	for(int j=0; j<t; j++) U[j]=0.0;

	//  组成法方程式 
	for(int k=0;k<n;k++)
	{
		for(int s=0;s<n;s++)
		{
			for(int i=0;i<t;i++)
			{
				U[i]+=A[k*t+i]*P[ij(k,s)]*L[s];
				for(int j=0; j<=i; j++)
				{
					N[ij(i,j)]+=A[k*t+i]*P[ij(k,s)]*A[s*t+j];
				}
			}
		}
	}
	
	//法方程系数矩阵求逆
	if(!inverse(N,t))
	{
		free(U);
		return -1.0;
	}

	//计算X
	for(int i=0;i<t;i++)
	{
		X[i]=0.0;
		for(int j=0;j<t;j++)X[i]+=N[ij(i,j)]*U[j];
	}
	free(U);
	
	//  计算V
	for(int i=0;i<n;i++)
	{
		double vi=-L[i];
		for(int j=0;j<t;j++) vi+=A[i*t+j]*X[j];
		V[i]=vi;
	}
	
	if(n==t)return 0.0;

	double pvv=Calculate_q(P,V,n);

	return sqrt(pvv/(n-t));
								
}







//////////////////////////////////////////////////////////////////////////
//         条件平差: Condition Adjustment
//         条件平差: Adjustment of Condition equations
// 参数：
// int n - 观测值个数
// int r - 条件个数
// double B[] - 条件方程系数矩阵，数组长度为r×n
// double W[] - 条件方程自由项向量，数组长度为n
// double P[] - 观测值权矩阵的对角线元素,数组长度为n
// double Q[] - 观测值平差值的权逆阵，仅存下三角矩阵，数组长度为n×(n+1)/2 
// double V[] - 残差向量，数组长度为n
// 返回值：验后单位权中误差
//////////////////////////////////////////////////////////////////////////
double Condition(int n, int r, double B[],double W[],double P[],
			   double Q[],double V[])
{
	double *N=(double *)malloc(sizeof(double)*r*(r+1)/2);//法方程系数矩阵
	
	// 法方程系数阵计算
	for(int i=0;i<r;i++)
	{
		for(int j=0;j<=i;j++)
		{
			double nij=0.0;
			for(int k=0;k<n;k++)
					nij+=B[i*n+k]*B[j*n+k]/P[k];
			N[ij(i,j)]=nij;
		}
	}
	
	// 系数矩阵求逆
	if(inverse(N,r)==false)
	{
		free(N);
		return -1.0;
	}

	double *K=(double *)malloc(sizeof(double)*r);
	for(int i=0; i<r; i++)
	{
		K[i]=0.0;
		for(int j=0;j<r;j++)K[i]-=N[ij(i,j)]*W[j];
	}
	
	//  计算V
	double pvv=0;
	for(int i=0;i<n;i++)
	{
		double vi=0.0;
		for(int j=0;j<r;j++) vi+=B[j*n+i]*K[j];
		vi/=P[i];
		V[i]=vi;
		pvv+=vi*vi*P[i];
	}

	//观测值平差值的权逆阵计算
	for(int i=0;i<n;i++)
		for(int j=0;j<=i;j++)
		{
			double qij = (i==j)? 1.0/P[i]: 0.0;
			for(int k=0;k<r;k++)
				for(int s=0;s<r;s++)
					qij-=B[k*n+i]*N[ij(k,s)]*B[s*n+j]/P[i]/P[j];
	
			Q[ij(i,j)]=qij;
		}

	free(N);
	free(K);
	return sqrt(pvv/r);
}






//////////////////////////////////////////////////////////////////////////
//     相关观测值的条件平差
// 参数：
// int n - 观测值个数
// int r - 条件个数
// double B[] - 条件方程系数矩阵，数组长度为r×n
// double L[] - 观测值向量,数组长度为n
// double Q[] - 函数调用前为观测值的权逆阵，调用后为观测值平差值的权逆阵，
//              仅存下三角矩阵，数组长度为n×(n+1)/2 
// double V[] - 残差向量，数组长度为n
// 返回值：验后单位权中误差
//////////////////////////////////////////////////////////////////////////
double Co_Condition(int n, int r, double B[],double W[],
					double Q[],double V[])
{
	int rr=r*(r+1)/2;

	double *N=(double *)malloc(sizeof(double)*rr); //临时数组，存放法方程系数阵
	for(int i=0; i<rr; i++) N[i]=0.0;
		
	double *M=(double *)malloc(sizeof(double)*r*n); //临时数组M，存放系数阵与权逆阵的乘积B*Q
	for(int i=0;i<r*n;i++)M[i]=0.0;

	//  计算M
	for(int i=0;i<r;i++)
		for(int j=0;j<n;j++)
			for(int k=0;k<n;k++)M[i*n+j]+=B[i*n+k]*Q[ij(k,j)];

	// 计算联系数法方程系数阵 N=M*B'
	int index=0;
	for(int i=0;i<r;i++)
		for(int j=0;j<=i;j++)
		{
			for(int k=0; k<n; k++)N[index]+=M[i*n+k]*B[j*n+k];
			index++;
		}

	//法矩阵求逆
	if(inverse(N,r)==false)
	{
		free(N);
		free(M);
		return -1.0;
	}
		
	double *K=(double *)malloc(sizeof(double)*r);
	double pvv=0.0;
	for(int i=0;i<r;i++)
	{
		double ki=0.0;
		for(int j=0;j<r;j++)ki-=N[ij(i,j)]*W[j];
		K[i]=ki;
		pvv-=ki*W[i];
	}
		
	//  计算V
	for(int i=0;i<n;i++)
	{
		double vi=0.0;
		for(int j=0;j<r;j++) vi+=M[j*n+i]*K[j];
		V[i]=vi;
	}
		
	//  计算QX
	index=0;
	for(int i=0;i<n;i++)
		for(int j=0;j<=i;j++)
		{
			for(int k=0;k<r;k++)
				for(int s=0;s<r;s++)
					Q[index]-=M[k*n+i]*N[ij(k,s)]*M[s*n+j];
			index++;
		}

	free(K);
	free(M);
		
	return sqrt(pvv/r);
			
}







//////////////////////////////////////////////////////////////////////////
//  具有条件的参数平差
//  parameter adjustment with constraint
double Adjust_With_Constraint(int n, int t,int r,bool correlative,
			double A[],double L[],double B[],double W[],double P[],
			double X[],double QX[],double V[])
{

	//  (1)参数平差
	double m; 
	if(correlative)m=Co_Adjust(n,t,A,L,P,X,QX,V);
	else m=Adjust(n,t,A,L,P,X,QX,V);
	if(m<0.0) return -1.0;
	
	//  (2)条件方程自由项更新 
	for(int i=0;i<r;i++)
	{
		for(int j=0;j<t;j++) 
		{
			W[i]+=B[i*t+j]*X[j];
		}
	}
		
	//  (3)条件平差
	double *dX=V;
	m=Co_Condition(t,r,B,W,QX,dX);
	if(m<0.0) return -1.0;

	for(int i=0;i<t;i++) X[i]+=dX[i];
	
	// (4) 计 算 V 
	double pvv=0;
	for(int i=0;i<n;i++)
	{
		double vi=-L[i];
		for(int j=0;j<t;j++) 
		{
			vi+=A[i*t+j]*X[j];
		}
		V[i]=vi;
		pvv+=V[i]*V[i]*P[i];
	}

	if(n-t+r==0)return 0.0; //无多余观测
	
	if(correlative) 
	{
		pvv=Calculate_q(P,V,n);
	}

	return sqrt(pvv/(n-t+r));
	
}







//////////////////////////////////////////////////////////////////////////
//    具有参数的条件平差:  Condition Adjustment with parameters
double Condition_With_X(int n, int r,int t,bool correlative,
						double B[],double A[],double W[],double P[],
						double X[],double QX[],double V[])
{
	double *Q=(double *)malloc(sizeof(double)*(n+1)*n/2); //观测值的权逆阵
	double *N=(double *)malloc(sizeof(double)*(r+1)*r/2); //法方程系数矩阵

	//     计算矩阵N
	if(correlative) //观测值相关
	{   
		for(int i=0; i<n*(n+1)/2; i++)Q[i]=P[i];
		if(inverse(Q,n)==false)
		{
			fprintf(stderr,"观测值的权矩阵奇异！\n");
			return -1.0;
		}

		for(int i=0; i<r; i++)
		{
			for(int j=0; j<=i; j++)
			{
				double nij=0.0;
				for(int k=0;k<n;k++)
				{
					for(int s=0;s<n;s++)
					{
						nij+=B[i*n+k]*Q[ij(k,s)]*B[j*n+s];
					}
				}
				N[ij(i,j)]=nij;
			}
		}
	}
	else  	// 观测值独立
	{	
		for(int i=0; i<r; i++)
		{
			for(int j=0; j<=i; j++)
			{
				double nij=0.0;
				for(int k=0;k<n;k++)
				{
					nij+=B[i*n+k]*B[j*n+k]/P[k];
				}
				N[ij(i,j)]=nij;
			}
		}
	}
	
	if(inverse(N,r)==false)
	{   
		free(Q);
		free(N);
		return -1.0;
	}

//	for(int i=0;i<r;i++) W[i]=-W[i];
	double m=Co_Adjust(r,t,A,W,N,X,QX,V);
	for(int i=0; i<t; i++) X[i]=-X[i];
	
	if(m<0.0)
	{
		free(Q);
		free(N);
		return -1.0;
	}

	double *K=(double *)malloc(sizeof(double)*r);
	for(int i=0; i<r; i++)
	{
		double ki=0.0;
		for(int j=0; j<r; j++)
		{
		   ki+=N[ij(i,j)]*V[j];
		}
		K[i]=ki;
	}

	//  计算向量：V
	double pvv=0.0;
	if(correlative) //观测值相关
	{   
		for(int i=0;i<n;i++)
		{
			double vi=0.0;
			for(int k=0;k<n;k++)
			{
				for(int s=0;s<r;s++)
				{
					vi+=Q[ij(i,k)]*B[s*n+k]*K[s];
				}
			}
			V[i]=vi;
		}
		pvv=Calculate_q(P,V,n);
	}		
	else //观测值独立
	{   
		for(int i=0;i<n;i++)
		{
			double vi=0.0;
			for(int s=0;s<r;s++)
			{
				vi+=B[s*n+i]*K[s]/P[i];
			}
			V[i]=vi;
			pvv+=P[i]*vi*vi;
		}
	}		
		
	free(Q);
	free(N);
	free(K);
	
	return sqrt(pvv/(r-t));
}





