// Collocation.cpp: implementation of the CCollocation class.
//
//////////////////////////////////////////////////////////////////////

#include "collocation.h"
#include "classic.h"
#include "modern.h"
#include "public.h"
#include <malloc.h>

//////////////////////////////////////////////////////////////////////////
/*  缺省的构造函数
CCollocation::CCollocation()
{
	n=0;   //观测值总数
	IsCompute = false;
}
*/
//////////////////////////////////////////////////////////////////////////
/*  析构函数
CCollocation::~CCollocation()
{
	if(n==0)return ;
	delete []PL;
	delete []QX;
	delete []X;
	delete []G;
}
*/
//////////////////////////////////////////////////////////////////////////
//初始化函数
void init_collocation(struct collocation *coll)
{
	coll->IsCompute=false;
	coll->n=0;
	coll->t=0;
	coll->A=NULL;
	coll->L=NULL;
	coll->P=NULL;
	coll->PL=NULL;
	coll->QS=NULL;
	coll->G=NULL;
	coll->IsCompute=false;
	coll->X=NULL;
	coll->QX=NULL;
	coll->CoCompute=&collocation_Compute;
	coll->CoGet_S1=&collocation_Get_S1;
	coll->CoGet_QS1=&collocation_Get_QS1;
	coll->CoGet_S2=&collocation_Get_S2;
	coll->CoGet_QS2=&collocation_Get_QS2;
	coll->CoGet_T=&collocation_Get_T;
	coll->CoGet_QT=&collocation_Get_QT;
	coll->CoGet_V=&collocation_Get_V;
}
//////////////////////////////////////////////////////////////////////////
// 析构函数  释放 结构中不为空的指针
void de_collocation(struct collocation *coll)
{
	if(coll->A) free(coll->A);
	if(coll->P) free(coll->P);
	if(coll->L) free(coll->L);
	if(coll->PL) free(coll->PL);
	if(coll->QS) free(coll->QS);bug!!!!!!
	if(coll->G) free(coll->G);
	if(coll->X) free(coll->X);
	if(coll->QX) free(coll->QX);
}

//////////////////////////////////////////////////////////////////////////
//  数组初始化、非随机参数计算函数
bool collocation_Compute(double *A_,double *L_,double *P_,double *QS_,int n_,int t_,
				struct collocation *coll)
{
	
	int n=(coll->n)=n_;
	int t=(coll->t)=t_;

	double *P=coll->P=P_;
	double *QS=coll->QS=QS_;
	double *A=coll->A=A_;
	
	double *L=coll->L=L_;
	
	
	//A=A_;   //误差方程系数阵
	//L=L_;   //观测值
	//P=P_;   //观测值的权矩阵，对称正定矩阵，仅存下三角元素
	//QS=QS_; //已测点信号的协方差矩阵，仅存下三角元素
	//n=n_;   //观测值总数
	//t=t_;   //参数总数
	
	
	
	coll->X=(double *)malloc(sizeof(double)*t); //参数估值
	coll->PL=(double *)malloc(sizeof(double)*n*(n+1)/2); //预测残差的权矩阵
	coll->QX=(double *)malloc(sizeof(double)*t*(t+1)/2); //X的权逆阵
	
	coll->G=(double *)malloc(sizeof(double)*n*(n+1)/2);; //临时矩阵,用于QS计算
	double *X=coll->X;
	double *QX=coll->QX;
	double *PL=coll->PL;
	double *G=coll->G;
	
	int nn=n*(n+1)/2;
	
	
	for(int i=0; i<nn; i++)PL[i]=P[i];
	
	if(!inverse(PL,n)) return false;
	
	for(int i=0;i<nn;i++) PL[i]=PL[i]+QS[i];
	
	if(!inverse(PL,n)) return false;
	
	double *U=(double *)malloc(sizeof(double)*t);
	double *N=coll->QX;
	
	//法方程系数阵和自由项清零
	for(int i=0; i<t*(t+1)/2; i++) N[i]=0.0;
	for(int i=0; i<t; i++) U[i]=0.0;
	
	//  组成法方程式 
	for(int k=0;k<n;k++)
	{
		for(int s=0;s<n;s++)
		{
			for(int i=0;i<t;i++)
			{
				U[i]+=A[k*t+i]*PL[ij(k,s)]*L[s];
				for(int j=0; j<=i; j++)
				{
					N[ij(i,j)]+=A[k*t+i]*PL[ij(k,s)]*A[s*t+j];
				}
			}
		}
	}
	
	//法方程系数矩阵求逆
	if(!inverse(N,t))
	{
		free(U);
		return false;
	}
	
	//计算X
	for(int i=0;i<t;i++)
	{
		X[i]=0.0;
		for(int j=0;j<t;j++)
		{
			X[i]+=N[ij(i,j)]*U[j];
		}
	}
	free(U);
	
	
	double *M=(double *)malloc(sizeof(double)*t*n); // P*A
	
	for(int i=0;i<n;i++)
	{
		for(int j=0;j<t;j++)
		{
			double mij=0.0;
			for(int k=0;k<n;k++)
				mij+=PL[ij(i,k)]*A[k*t+j];
			M[i*t+j]=mij;
		}
	}
	
	for(int i=0; i<nn; i++) G[i]=0.0;
	Calculate_BQBT(M,QX,n,t,G);
	for(int i=0; i<nn; i++) G[i]=PL[i]-G[i];
	
	free(M);
	
	coll->IsCompute=true;
	return true;
}	


//////////////////////////////////////////////////////////////////////
//    已测点信号估值计算
bool collocation_Get_S1(double *S,struct collocation *coll)//S的大小？？？安全？
{
	int n=(coll->n);
	int t=(coll->t);
	double *PL=coll->PL;
	double *P=coll->P;
	double *QS=coll->QS;
	double *A=coll->A;
	double *G=coll->G;
	double *L=coll->L;
	double *X=coll->X;
	double *QX=coll->QX;
	if(!coll->IsCompute) return false;
	
	for(int i=0; i< coll->n; i++)
	{
		S[i]=0.0;
		for(int j=0; j< coll->n; j++)
		{
			double Vj=-(coll->L[j]);
			for(int k=0; k< coll->t; k++)
			{
				Vj+=(coll->A[j*t+k])*(coll->X[k]);
			}
			for(int k=0; k<n; k++)
			{
				S[i]-=(coll->QS[ij(i,k)])*(coll->PL[ij(k,j)])*Vj;
			}
		}
	}

	return true;
}

//////////////////////////////////////////////////////////////////////////
//   未测点信号的估值
bool collocation_Get_S2(double *S2,double *QS21,int t2,struct collocation *coll) //安全？？？
{
	int n=(coll->n);
	int t=(coll->t);
	double *PL=coll->PL;
	double *P=coll->P;
	double *QS=coll->QS;
	double *A=coll->A;
	double *G=coll->G;
	double *L=coll->L;
	double *X=coll->X;
	double *QX=coll->QX;
	if(!coll->IsCompute) return false;
	
	for(int i=0; i<t2; i++)
	{
		S2[i]=0.0;
		
		for(int j=0;j<(coll->n);j++)
		{
			double Vj=-(coll->L[j]);
			for(int k=0; k<t; k++)
			{
				Vj+=(coll->A[j*t+k])*(coll->X[k]);
			}
			
			for(int k=0;k<n;k++)
			{
				S2[i]-=QS21[i*n+k]*(coll->PL[ij(k,j)])*Vj;
			}
		}
	}
	
	return true;
}

//////////////////////////////////////////////////////////////////////////
//   已测点信号估值的权逆阵
bool collocation_Get_QS1(double *QQ,struct collocation *coll)
{
	int n=(coll->n);
	int t=(coll->t);
	double *PL=coll->PL;
	double *P=coll->P;
	double *QS=coll->QS;
	double *A=coll->A;
	double *G=coll->G;
	double *L=coll->L;
	double *X=coll->X;
	double *QX=coll->QX;
	if(!coll->IsCompute) return false;
	
	for(int i=0; i<n*(n+1)/2;i++) QQ[i]=0.0;
	
	for(int i=0; i<n; i++)
	{
		for(int j=0;j<=i;j++)
		{
			double qq=0.0;
			for(int k=0;k<n;k++)
				for(int s=0;s<n;s++)
					qq+=(coll->QS[ij(i,k)])*(coll->G[ij(k,s)])*(coll->QS[ij(s,j)]);
				QQ[ij(i,j)]=(coll->QS[ij(i,j)])-qq;
		}
	}
	
	return true;
}


//////////////////////////////////////////////////////////////////////////
//   未测点信号的权逆阵
bool collocation_Get_QS2(double *QS2,double *QS21,double *QQ,int t2,struct collocation *coll)
{
	if(!coll->IsCompute) return false;
	
	
	for(int i=0; i<t2*(t2+1)/2;i++) QQ[i]=0.0;
	
	Calculate_BQBT(QS21,coll->G,t2,coll->n,QQ);
	for(int i=0; i<t2*(t2+1)/2;i++) QQ[i]=QS2[i]-QQ[i];
	
	return true;
}



//////////////////////////////////////////////////////////////////////////
//   完全信号的估值
bool collocation_Get_T(double *T,double *QS21,double *A1,int t2,struct collocation *coll)
{
	int n=(coll->n);
	int t=(coll->t);
	double *PL=coll->PL;
	double *P=coll->P;
	double *QS=coll->QS;
	double *A=coll->A;
	double *G=coll->G;
	double *L=coll->L;
	double *X=coll->X;
	double *QX=coll->QX;
	if(!coll->IsCompute) return false;
	
	coll->CoGet_S2(T,QS21,t2,coll);

	for(int i=0;i<t2;i++)
	{
		for(int k=0;k<t;k++)
			T[i]+=A1[i*t+k]*(coll->X[k]);
	}
	
	return true;
}

//////////////////////////////////////////////////////////////////////////
//   完全信号的权逆阵
bool collocation_Get_QT(double *QS2,double *QS21,double *A1,double *QQ,int t2,struct collocation *coll)
{
	int n=(coll->n);
	int t=(coll->t);
	double *PL=coll->PL;
	double *P=coll->P;
	double *QS=coll->QS;
	double *A=coll->A;
	double *G=coll->G;
	double *L=coll->L;
	double *X=coll->X;
	double *QX=coll->QX;
	if(!coll->IsCompute) return false;
	for(int i=0; i<t2;i++) 
	{
		for(int j=0;j<=i;j++)
		{
			double qq=0.0;
			for(int k=0;k<n;k++)
				for(int s=0;s<n;s++)
					qq+=QS21[i*n+k]*(coll->PL[ij(k,s)])*QS21[j*n+s];
			QQ[ij(i,j)]=QS2[ij(i,j)]-qq;
		}
	}

	double *M=(double *)malloc(sizeof(double)*t2*t);
	for(int i=0; i<t2;i++) 
	{
		for(int j=0;j<t;j++)
		{
			double qq=0.0;
			for(int k=0;k<n;k++)
				for(int s=0;s<n;s++)
					qq+=QS21[i*n+k]*(coll->PL[ij(k,s)])*(coll->A[s*t+j]);
			M[i*t+j]=qq-A1[i*t+j];
		}
	}
	
	Calculate_BQBT(M,QX,t2,t,QQ);
	return true;
	
}


//////////////////////////////////////////////////////////////////////
//    已测点观测值残差
bool collocation_Get_V(double *V,struct collocation *coll)
{
	int n=(coll->n);
	int t=(coll->t);
	double *PL=coll->PL;
	double *P=coll->P;
	double *QS=coll->QS;
	double *A=coll->A;
	double *G=coll->G;
	double *L=coll->L;
	double *X=coll->X;
	double *QX=coll->QX;
	if(!coll->IsCompute) return false;
	
	if(!coll->CoGet_S1(V,coll)) return false; //计算已测点信号估值
	

	for(int i=0; i<n; i++)
	{
		V[i]-=coll->L[i];
		for(int j=0; j<t; j++)
		{
			V[i]+=(coll->A[i*t+j])*(coll->X[j]);
		}
	}
	
	return true;
}
	
	


