// collocation.h  最小二乘配置
//
//////////////////////////////////////////////////////////////////////
#ifndef _COLLOCATION_H
#define _COLLOCATION_H

#include <stdbool.h>

struct collocation{
	int n;//观测值总数和非随机参数总数
	int t;
	
	double *A;//误差方程系数矩阵，观测值向量，权矩阵
	double *L;
	double *P;
	
	double *PL; //自由项的权矩阵
	
	double *QS; //显信号的权逆阵，临时数组
	double *G;
	
	bool IsCompute;  // 是否调用过Compute函数
	
	double *X; // 系统性参数向量估值及其权逆阵
	double *QX;
	
	//  数组初始化、非随机参数计算函数
	bool (*CoCompute)(double *A_,double *L_,double *P_,double *QS_,int n_,int t_,
				struct collocation *coll);
	
		
	//计算显信号及其协方差
	bool (*CoGet_S1)(double *S,struct collocation *coll);
	bool (*CoGet_QS1)(double *QQ,struct collocation *coll); 

	//计算隐信号估值及其协方差
	bool (*CoGet_S2)(double *S2,double *QS21,int t2,struct collocation *coll); 
	bool (*CoGet_QS2)(double *QS2,double *QS21,double *QQ,int t2,struct collocation *coll);
	
	//计算完全信号及其协方差
	bool (*CoGet_T)(double *T,double *QS21,double *A1,int t2,struct collocation *coll);
	bool (*CoGet_QT)(double *QS2,double *QS21,double *A1,double *QQ,int t2,struct collocation *coll);

	// 计算残差
	bool (*CoGet_V)(double *V,struct collocation *coll);
	
};

//  数组初始化、非随机参数计算函数
bool collocation_Compute(double *A_,double *L_,double *P_,double *QS_,int n_,int t_,
				struct collocation *coll); 
	
//计算显信号及其协方差
bool collocation_Get_S1(double *S,struct collocation *coll);  //S的大小？？？安全？
bool collocation_Get_QS1(double *QQ,struct collocation *coll); 

//计算隐信号估值及其协方差
bool collocation_Get_S2(double *S2,double *QS21,int t2,struct collocation *coll); 
bool collocation_Get_QS2(double *QS2,double *QS21,double *QQ,int t2,struct collocation *coll);
	
//计算完全信号及其协方差
bool collocation_Get_T(double *T,double *QS21,double *A1,int t2,struct collocation *coll);
bool collocation_Get_QT(double *QS2,double *QS21,double *A1,double *QQ,int t2,struct collocation *coll);

// 计算残差
bool collocation_Get_V(double *V,struct collocation *coll);


///////////////////////////////////////////
//！！！！这些东西很危险，在多线程时	
//double *X,*QX; // 系统性参数向量估值及其权逆阵
	
//int n,t; //观测值总数和非随机参数总数
//double *A,*L,*P; //误差方程系数矩阵，观测值向量，权矩阵
//double *PL;   //自由项的权矩阵，
//double *QS,*G; //显信号的权逆阵，临时数组
		
//bool IsCompute;  // 是否调用过Compute函数



void init_collocation(struct collocation *coll);  //初始化函数
void de_collocation(struct collocation *coll);  //析构函数

void collocation_example();
void collocation_example2();


#endif // _COLLOCATION_H
