// LevelingAdjust.cpp: implementation of the CLevelingAdjust class.
//
//////////////////////////////////////////////////////////////////////

#include <stdlib.h>
#include <stdarg.h>
#include "probability.h"

#include "levelingad.h"
#include <math.h>
#include <string.h>
#include "public.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

void CLevelingAdjust_CLevelingAdjust(struct CLevelingAdjust *cla)
{
	cla->m_Lnumber=0;
	cla->m_Pnumber=0;
	cla->m_StablePnumber=0;
	
	cla->CLevelingAdjust=CLevelingAdjust_CLevelingAdjust;
	cla->de_CLevelingAdjust=CLevelingAdjust_de_CLevelingAdjust;
	
	cla->Inputdata=CLevelingAdjust_Inputdata;
	cla->Printdata=CLevelingAdjust_Printdata;
	cla->GetStationNumber=CLevelingAdjust_GetStationNumber;
	cla->ca_H0=CLevelingAdjust_ca_H0;
	cla->ca_ATPA=CLevelingAdjust_ca_ATPA;
	cla->ca_dX=CLevelingAdjust_ca_dX;
	cla->PrintResult=CLevelingdjust_PrintResult;
	cla->ca_V=CLevelingAdjust_ca_V;
	cla->LS_Adjustment=CLevelingAdjust_LS_Adjustment;
	cla->Quasi_Stable=CLevelingAdjust_Quasi_Stable;
	cla->FreeNetAdjust=CLevelingAdjust_FreeNetAdjust;
	cla->FindShortPath=CLevelingAdjust_FindShortPath;
	cla->LineClosure=CLevelingAdjust_LineClosure;
	cla->LoopClosure=CLevelingAdjust_LoopClosure;
	cla->DataSnooping=CLevelingAdjust_DataSnooping;
}

void CLevelingAdjust_de_CLevelingAdjust(struct CLevelingAdjust *cla)
{
	if(cla->m_Lnumber>0)
	{
		free(cla->StartP);  
		free(cla->EndP);    
		free(cla->L);       
		free(cla->P);       
		free(cla->V);       
	}

	if(cla->m_Pnumber>0)
	{
		free(cla->Height);   
		free(cla->ATPA);     
		free(cla->ATPL);     
		free(cla->dX);       
		
		for(int i=0; i<cla->m_Pnumber;i++)
			if(cla->Pname[i]!=NULL)free(cla->Pname[i]);
		free(cla->Pname);    
	}

}



//////////////////////////////////////////////////////////////////////////
//   原始数据输入函数
void CLevelingAdjust_Inputdata(char *datafile,struct CLevelingAdjust *cla)
{
     FILE *fp;
     if((fp=fopen(datafile,"r"))==NULL)
     {	       
		 fprintf(stderr,"\n 数据文件打不开!\n");  
		 exit(0);
	 }

     fscanf(fp,"%d%d%d",&(cla->m_Lnumber),&(cla->m_Pnumber),&(cla->m_knPnumber));
     
	 int unPnumber=cla->m_Pnumber-cla->m_knPnumber;
	
     cla->Height=(double *)malloc(sizeof(double)*cla->m_Pnumber);
     cla->dX=(double *)malloc(sizeof(double)*cla->m_Pnumber);
     cla->ATPA=(double *)malloc(sizeof(double)*cla->m_Pnumber*(cla->m_Pnumber+1)/2);
     cla->ATPL=(double *)malloc(sizeof(double)*cla->m_Pnumber);
     cla->StartP=(int *)malloc(sizeof(int)*cla->m_Lnumber);
     cla->EndP=(int *)malloc(sizeof(int)*cla->m_Lnumber);
     cla->L=(double *)malloc(sizeof(double)*cla->m_Lnumber);
     cla->V=(double *)malloc(sizeof(double)*cla->m_Lnumber);
     cla->P=(double *)malloc(sizeof(double)*cla->m_Lnumber);
     fscanf(fp,"%lf",&(cla->m_Sigma));
	
     cla->Pname=(char **)malloc(sizeof(char *)*cla->m_Pnumber);
	 for(int i=0;i<cla->m_Pnumber;i++)
	 {
		 // GetStationNumber函数根据Pname[i]是否为NULL
		 // 确定Pname[i]是否为点名地址
		 cla->Pname[i] = NULL;
	 }

	 char buffer[100]; //临时数组，保存从文件中读到的点名	 
	 
	 //  读取已知高程数据
     for(int i=0;i<=cla->m_knPnumber-1;i++)
     {
		   fscanf(fp,"%s",buffer);
		   int c=cla->GetStationNumber(buffer,cla);
		   
		   fscanf(fp,"%lf",&(cla->Height[c]));
     }

	 //  读取观测数据
     for(int i=0;i<cla->m_Lnumber;i++)
     {
		 fscanf(fp,"%s",buffer); //读取高程起点名
	 	 cla->StartP[i]=cla->GetStationNumber(buffer,cla); 
		 if(cla->StartP[i]<0)
		 {
			 fprintf(cla->resultfp,"\n数据文件出错：");
			 fprintf(cla->resultfp,"\n第%d个高差的起始点名为\"%s\"",i+1,buffer);
			 fclose(cla->resultfp);
			 exit(0);
		 }
		 
		 fscanf(fp,"%s",buffer);//读取高程终点
		 cla->EndP[i]=cla->GetStationNumber(buffer,cla);

		 if(cla->EndP[i]<0)
		 {
			 fprintf(cla->resultfp,"\n数据文件出错：");
			 fprintf(cla->resultfp,"\n第%d个高差终点的点名为\"%s\"",i+1,buffer);
			 fclose(cla->resultfp);
			 exit(0);
		 }		 
		 
		 fscanf(fp,"%lf%lf",&cla->L[i],&cla->P[i]); //读取高差值与路线长度
		 cla->P[i]=1.0/cla->P[i]; 
     }

     fclose(fp);
}

//////////////////////////////////////////////////////////////////////////
//   原始数据写到结果文件
void CLevelingAdjust_Printdata(struct CLevelingAdjust *cla)
{
     int i;
     fprintf(cla->resultfp,"\n 观测值总数: %d  总点数: %d  已知点数：%d \n",
						cla->m_Lnumber, cla->m_Pnumber,cla->m_knPnumber);
     fprintf(cla->resultfp,"\n 验前单位权中误差：%lf" ,cla->m_Sigma);
     fprintf(cla->resultfp,"\n\n   已知高程:\n");
     for(i=0;i<=cla->m_knPnumber-1;i++)
     {
		   fprintf(cla->resultfp,"\n%5d %8s  ",i+1,cla->Pname[i]);
		   fprintf(cla->resultfp,"%10.4lf  ",cla->Height[i]);
      }

     fprintf(cla->resultfp,"\n\n   高差观测值:\n");
     for(i=0;i<=cla->m_Lnumber-1;i++)
     {
		   fprintf(cla->resultfp,"\n%5d %8s %8s",i+1,cla->Pname[cla->StartP[i]],cla->Pname[cla->EndP[i]]);
		   fprintf(cla->resultfp,"%12.4lf %10.3lf",cla->L[i],1.0/cla->P[i]);
     }
}


//////////////////////////////////////////////////////////////////////////
//	 点名存贮，返回点名对应的点号
int  CLevelingAdjust_GetStationNumber(char *name,struct CLevelingAdjust *cla)
{
    for(int i=0; i<cla->m_Pnumber; i++) 
	{
		if(cla->Pname[i]!=NULL)
		{
			//将待查点名与已经存入点名数组的点名比较
			if(strcmp(name,cla->Pname[i])==0)return i;
		}
		else
		{
			//待查点名是一个新的点名，将新点名的地址放到Pname数组中
			int len = strlen(name);
			cla->Pname[i] = (char *)malloc(sizeof(char)*(len+1));
			strcpy(cla->Pname[i], name);
			return i;
		}
	}

	return -1;  //Pname数组已经存满，且没有待查点名
    
}                                          




                                           
//////////////////////////////////////////////////////////////////////////
//	      高程近似值计算
void CLevelingAdjust_ca_H0(struct CLevelingAdjust *cla)
{
     for(int i=cla->m_knPnumber;i<cla->m_Pnumber;i++) cla->Height[i]=-9999.9;
     
	 int jj=0; //计算出近似高程的点数
	 for(int ii=1;;ii++)
     {
		 for(int i=0;i<cla->m_Lnumber;i++)
		 {
			 int k1=cla->StartP[i];  //高差起点号
			 int k2=cla->EndP[i];    //高差终点号
			 if(cla->Height[k1]>-9999.0 && cla->Height[k2]<-9999.0)
			 {
				 cla->Height[k2]=cla->Height[k1]+cla->L[i];
				 jj++;
			 }
			 else if(cla->Height[k1]<-9999.0 && cla->Height[k2]>-9999.0)
			 {
				 cla->Height[k1]=cla->Height[k2]-cla->L[i];
				 jj++;
			 }
		 } 
	       
		 if(jj==(cla->m_Pnumber-cla->m_knPnumber))break;
	     if(ii>(cla->m_Pnumber-cla->m_knPnumber))
	     {
			 fprintf(cla->resultfp,"\n\n下列点无法计算出概略高程:\n");
	    	 for(int i=0;i<cla->m_Pnumber;i++)
			 {
				 if(cla->Height[i]<-9999.0) printf("\n%s",cla->Pname[i]);
			 }
			 fprintf(stderr,"近似高程计算失败！\n");
			 fclose(cla->resultfp);
			 exit(0);
	     }
     } 
}


//////////////////////////////////////////////////////////////////////////
//  组成法方程
void CLevelingAdjust_ca_ATPA(struct CLevelingAdjust *cla)
{
    int t=cla->m_Pnumber;

    for(int i=0; i<t*(t+1)/2; i++) cla->ATPA[i]=0.0;
    for(int i=0; i<t; i++) cla->ATPL[i]=0.0;
    
    for(int k=0; k<cla->m_Lnumber; k++)
    {
       	int i=cla->StartP[k];
		int j=cla->EndP[k];
       	double Pk=cla->P[k];
       	double Lk=cla->L[k]-(cla->Height[j]-cla->Height[i]);

       	cla->ATPL[i]-=Pk*Lk;
	    cla->ATPL[j]+=Pk*Lk;
		cla->ATPA[ij(i,i)]+=Pk;
	    cla->ATPA[ij(j,j)]+=Pk;
	    cla->ATPA[ij(i,j)]-=Pk;
     }
}

//////////////////////////////////////////////////////////////////////////
//    高程平差值计算
void CLevelingAdjust_ca_dX(struct CLevelingAdjust *cla)
{
     if(!inverse(cla->ATPA,cla->m_Pnumber))
     {
          fprintf(stderr,"\n法方程系数矩阵降秩!\n");
		  fclose(cla->resultfp);
		  exit(0);
     }

     for(int i=0; i<cla->m_Pnumber; i++)
     {
		 double xi=0.0;
	     for(int j=0; j<cla->m_Pnumber; j++)
		 {
			 xi+=cla->ATPA[ij(i,j)]*cla->ATPL[j];
		 }
	     cla->dX[i]=xi;
		 cla->Height[i]+=xi;
     }
}


//////////////////////////////////////////////////////////////////////////
//   残差计算
double  CLevelingAdjust_ca_V(struct CLevelingAdjust *cla)
{
	double pvv=0.0;
	for(int i=0;i<=cla->m_Lnumber-1;i++)
	{
		 int k1=cla->StartP[i];
		 int k2=cla->EndP[i];
		 cla->V[i]=cla->Height[k2]-cla->Height[k1]-cla->L[i];
		 pvv+=cla->V[i]*cla->V[i]*cla->P[i];
	}
	return(pvv);
}

//////////////////////////////////////////////////////////////////////////
//   平差值输出
void  CLevelingdjust_PrintResult(struct CLevelingAdjust *cla)
{
	fprintf(cla->resultfp,"\n\n           ==== 高程平差值及其精度 ====\n");
	fprintf(cla->resultfp,"\n    点名   近似高程   改正数   高程平差值  中误差\n");
	for(int i=0; i<cla->m_Pnumber; i++)
	{
		fprintf(cla->resultfp,"\n %5s ",cla->Pname[i]);
	    double dx=cla->dX[i];
	    double qii=cla->ATPA[ij(i,i)];
	    fprintf(cla->resultfp,"%12.4lf%9.4lf%12.4lf%9.4lf",
					  cla->Height[i]-dx,dx,cla->Height[i],sqrt(qii)*cla->m_mu);
	}
	fprintf(cla->resultfp,"\n\n\n              ==== 观测值平差值及其精度 ====\n");
	fprintf(cla->resultfp,"\n   No. 起 点  终 点  观测高差    ｖ");
	fprintf(cla->resultfp,"     高差平差值   观测权   中误差\n");

	for(int i=0;i<=cla->m_Lnumber-1;i++)
	{
		int k1=cla->StartP[i];
	    int k2=cla->EndP[i];
	    double qii=cla->ATPA[ij(k1,k1)];
	    double qjj= cla->ATPA[ij(k2,k2)] ;
	    double qij=cla->ATPA[ij(k1,k2)];
	    double ml=sqrt(qii+qjj-2.0*qij)*cla->m_mu;
	    fprintf(cla->resultfp,"\n%5d %5s%5s ",i+1,cla->Pname[k1],cla->Pname[k2]);
	    fprintf(cla->resultfp,"%12.4lf %8.4lf%11.4lf%9.2lf%10.4lf",
				  cla->L[i],cla->V[i],cla->L[i]+cla->V[i],cla->P[i],ml);
	}
}

//////////////////////////////////////////////////////////////////////////
//    水准网最小二乘平差
void  CLevelingAdjust_LS_Adjustment(struct CLevelingAdjust *cla)
{
	cla->ca_H0(cla); //近似高程计算
	cla->ca_ATPA(cla); // 组成法方程
	
	//处理已知点
	for(int i=0;i<cla->m_knPnumber;i++) cla->ATPA[ij(i,i)]=1.0e30;
	
	cla->ca_dX(cla); // 高程平差值计算
	double pvv=cla->ca_V(cla);  // 残差计算
	
	cla->m_mu=sqrt(pvv/(cla->m_Lnumber-(cla->m_Pnumber-cla->m_knPnumber)));
	fprintf(cla->resultfp,"\n\n  最小二乘平差结果：\n    [pvv]=%-12.8lf",pvv);
	fprintf(cla->resultfp,"\n       μ=±%lf",cla->m_mu);

	cla->PrintResult(cla);
	return ;
}

//////////////////////////////////////////////////////////////////////////
//       水准网－自由网平差
void CLevelingAdjust_FreeNetAdjust(struct CLevelingAdjust *cla)
{
	cla->ca_ATPA(cla); //组成法方程
	
	int tt=cla->m_Pnumber*(cla->m_Pnumber+1)/2;
	for(int i=0; i<tt; i++)	cla->ATPA[i]+=1.0/cla->m_Pnumber;  

	cla->ca_dX(cla);

	for(int i=0;  i<tt; i++) cla->ATPA[i]-=1.0/cla->m_Pnumber; 
	
	double pvv = cla->ca_V(cla);
	fprintf(cla->resultfp,"\n    [pvv]=%-12.8lf",pvv);
	cla->m_mu= sqrt(pvv/(cla->m_Lnumber-(cla->m_Pnumber-1)));
	fprintf(cla->resultfp,"\n       μ=±%lf",cla->m_mu);
	
	cla->PrintResult(cla);
	
	PrintM2(cla->resultfp,cla->ATPA,cla->m_Pnumber,4,"%15e ","Qx",true);	
} 



//////////////////////////////////////////////////////////////////////////
//       水准网拟稳平差
//       2008－07－16
void CLevelingAdjust_Quasi_Stable(char *file,struct CLevelingAdjust *cla) //file是拟稳点名文件
{
	//打开文件，获取拟稳点名
	FILE *fp=fopen(file,"r");
	if(fp==NULL)
	{
		fprintf(stderr,"拟稳点文件打不开！\n");
		fclose(cla->resultfp);
		exit(0);
	}
	
	fscanf(fp,"%d",&cla->m_StablePnumber);
	if(cla->m_StablePnumber>cla->m_Pnumber || cla->m_StablePnumber<1)
	{
		fprintf(stderr,"拟稳点数错误！\n");
		fclose(cla->resultfp);
		exit(0);		
	}

	cla->IsStable=(int *)malloc(sizeof(int)*cla->m_Pnumber);
	
	for(int i=0;i<=cla->m_Pnumber-1;i++) cla->IsStable[i]=false;
		
	fprintf(cla->resultfp,"\n%s\n","\n   拟稳点：");
	for(int i=0;i<=cla->m_StablePnumber-1;i++)
	{
		char name[100];
		fscanf(fp,"%s",name);
		int k=cla->GetStationNumber(name,cla);
		if(k<0)
		{
			fprintf(cla->resultfp,"\n拟稳点名文件出错：");
			fprintf(cla->resultfp,"\n出现了数据文件中没有的点名：%s",name);
			fclose(cla->resultfp);
			exit(0);			
		}
		cla->IsStable[k] = 1;
		fprintf(cla->resultfp,"\n%10s ",cla->Pname[k]);
	}
	fclose(fp);

	cla->ca_ATPA(cla);
	for(int i=0;i<cla->m_Pnumber;i++)
	{
		for(int j=0;j<=i;j++)
			if(cla->IsStable[i] && cla->IsStable[j] )
				cla->ATPA[ij(i,j)]+=1.0/cla->m_StablePnumber;
	}
	
	cla->ca_dX(cla);
	
	for(int i=0;i<cla->m_Pnumber;i++)
	{
		for(int j=0;j<=i;j++)
			cla->ATPA[ij(i,j)]-=1.0/cla->m_StablePnumber;
	}

	double pvv = cla->ca_V(cla);
	cla->m_mu= sqrt(pvv/(cla->m_Lnumber-(cla->m_Pnumber-1)));
	
	fprintf(cla->resultfp,"\n    [pvv]=%-12.8lf",pvv);
	fprintf(cla->resultfp,"\n    单位权中误差:   m=%lf",cla->m_mu);
	
	cla->PrintResult(cla);
	PrintM2(cla->resultfp,cla->ATPA,cla->m_Pnumber,4,"%15e ","Qx",true);		
}


	
//////////////////////////////////////////////////////////////////////////
//	路线闭合差计算	
void CLevelingAdjust_LineClosure(struct CLevelingAdjust *cla)
{
	if(cla->m_knPnumber<2)return; // 已知点数小于2

	fprintf(cla->resultfp,"\n\n  === 路线闭合差计算 === ");
				
	int *neighbor=(int *)malloc(sizeof(int)*cla->m_Pnumber); //邻接点数组
	double *diff=(double *)malloc(sizeof(double)*cla->m_Pnumber); //高差累加值数组
	double *S=(double *)malloc(sizeof(double)*cla->m_Pnumber);    //路线长累加值数组
	
	for(int i=0; i<cla->m_knPnumber-1; i++)
	{
		cla->FindShortPath(i,-1,neighbor,diff,S,cla); //搜索最短路线，用所有观测值
		
		for(int j=i+1;j<cla->m_knPnumber;j++)
		{
			if(neighbor[j]<0)
			{
				fprintf(cla->resultfp,"%s-%s之间找到不到最短路线",
					cla->Pname[i],cla->Pname[j]);
				continue;
			}

			// 输出附合路线上的点号
			fprintf(cla->resultfp,"\n    附合路线：");
			int k=j;
			while(1)
			{
				fprintf(cla->resultfp,"%s - ",cla->Pname[k]);
				k=neighbor[k];
				if(k==i)break;
			}
			fprintf(cla->resultfp,"%s ",cla->Pname[i]);

			// 闭合差计算，限差计算与输出
			double W=cla->Height[i]+diff[j]-cla->Height[j]; // 闭合差
			double limit=2.0*sqrt(S[j])*cla->m_Sigma;  // 限差
			
			fprintf(cla->resultfp,"\n      W=%7.3lf  (限差:%7.3lf) ",
				-W,limit); // 因为输出的点号是反向的，所以闭合差为负号
		}
	}
	free(neighbor);
	free(S);
	free(diff);	
}

//////////////////////////////////////////////////////////////////////////
//  搜索最短路径
void CLevelingAdjust_FindShortPath(int p,int exclude,int neighbor[],
				double diff[],double S[],struct CLevelingAdjust *cla)
{
	for(int i=0; i<cla->m_Pnumber; i++) 
	{
		neighbor[i]=-1;  // 还没有邻接点
		S[i]=1.0e30;
	}
	S[p]=0.0;
	diff[p]=0.0;
	neighbor[p]=p;
	
	for(int i=0;;i++)
	{
		bool successful=true; 
		for(int j=0;j<=cla->m_Lnumber-1;j++)
		{
			if(j==exclude) continue;
			int p1=cla->StartP[j];
			int p2=cla->EndP[j];
			double S12=1.0/cla->P[j];
			if(neighbor[p1]<0 && neighbor[p2]<0) continue;
			
			if(S[p2]>S[p1]+S12)
			{
				neighbor[p2]=p1;
				S[p2]=S[p1]+S12;
				diff[p2]=diff[p1]+cla->L[j];
				successful=false;
			}
			else if(S[p1]>S[p2]+S12)
			{
				neighbor[p1]=p2;
				S[p1]=S[p2]+S12;
				diff[p1]=diff[p2]-cla->L[j];
				successful = false;
			}
		}
		if(successful)break;
	}

	return ;
}




//////////////////////////////////////////////////////////////////////////
//	环闭合差计算	
void CLevelingAdjust_LoopClosure(struct CLevelingAdjust *cla)
{		
	fprintf(cla->resultfp,"\n\n  === 环闭合差计算 === ");

	int num=cla->m_Lnumber-(cla->m_Pnumber-1);
	if(num<1)
	{
		fprintf(cla->resultfp,"\n    无闭合环\n\n");		
		return;
	}
	
	//int *neighbor=new int[m_Pnumber]; //邻接点数组
	//int *used=new int[m_Lnumber]; //观测值是否已经用于闭合差计算
	//double *diff=new double[m_Pnumber]; //高差累加值数组
	//double *S=new double[m_Pnumber];	//路线长数组
	int *neighbor=(int *)malloc(sizeof(int)*cla->m_Pnumber); //邻接点数组
	int *used=(int *)malloc(sizeof(int)*cla->m_Lnumber); //观测值是否已经用于闭合差计算
	double *diff=(double *)malloc(sizeof(double)*cla->m_Pnumber); //高差累加值数组
	double *S=(double *)malloc(sizeof(double)*cla->m_Pnumber);    //路线长数组
	
	for(int i=0; i<=cla->m_Lnumber-1; i++) used[i]=0;

	for(int i=0; i<cla->m_Lnumber; i++)
	{
		int k1=cla->StartP[i];
		int k2=cla->EndP[i];
		if(used[i]) continue;

		cla->FindShortPath(k2,i,neighbor,diff,S,cla);//搜索最短路线，第i号观测值不能参加
			
		if(neighbor[k1]<0)
		{
			fprintf(cla->resultfp,"\n观测值%s - %s与任何观测边不构成闭合环",
				cla->Pname[k1],cla->Pname[k2]);
		}
		else
		{
			used[i]=1;
			fprintf(cla->resultfp,"\n   环闭合差： ");
			int p1=k1;
			while(1)//输出点名
			{
				int p2=neighbor[p1];
				fprintf(cla->resultfp,"%s - ",cla->Pname[p1]);
			
				for(int r=0;r<cla->m_Lnumber;r++)//将用过的观测值标定
				{
					if(cla->StartP[r]==p1 && cla->EndP[r]==p2)
					{
						used[r]=1;
						break;
					}
					else if(cla->StartP[r]==p2 && cla->EndP[r]==p1)
					{
						used[r]=1;
						break;
					}
				}
				if(p2==k2)break;
				else p1=p2;
			}
			fprintf(cla->resultfp,"%s - %s  ",cla->Pname[k2],cla->Pname[k1]);

			double W=cla->L[i]+diff[k1];   //闭合差
			double SS=S[k1]+1.0/cla->P[i]; //环长
			fprintf(cla->resultfp,"\n     W=%7.4lf  (限差:%7.4lf) ",
				-W,2.0*sqrt(SS)*cla->m_Sigma);
		}
	}

	free(neighbor);
	free(used);
	free(diff);
	free(S);		
} 


#include "probability.h"
//////////////////////////////////////////////////////////////////////////
//   水准网粗差探测
void CLevelingAdjust_DataSnooping(double arfa,struct CLevelingAdjust *cla)
{
	//CProbability pp;
	double U=probability_re_norm(1.0-arfa/2.0);
	
	cla->ca_H0(cla); //近似高程计算	

	double pvv;
	int k=0;
    	for(k=0; ; k++)
	{
		 cla->ca_ATPA(cla); // 组成法方程
		 
		 //处理已知点
	     for(int i=0;i<cla->m_knPnumber;i++) cla->ATPA[ij(i,i)]=1.0e30;
	
	     cla->ca_dX(cla); // 高程平差值计算
	     pvv=cla->ca_V(cla); // 残差计算

		 double max_v=0.0;
		 int    max_i;
		 for(int i=0;i<=cla->m_Lnumber-1;i++)
		 {
			 if(cla->P[i]<1.0e-10) continue;
			 int k1=cla->StartP[i];
			 int k2=cla->EndP[i];
			 double qii=cla->ATPA[ij(k1,k1)];
			 double qjj= cla->ATPA[ij(k2,k2)] ;
			 double qij=cla->ATPA[ij(k1,k2)];
			 
			 double qv=1.0/cla->P[i]-(qii+qjj-2.0*qij);
			 double mv=sqrt(qv)*cla->m_Sigma;

			 double vi=cla->V[i]/mv;

			 if(fabs(vi) > max_v )
			 {
				 max_v=fabs(vi);
				 max_i=i;
			 }
		 }

		 if(max_v>U)cla->P[max_i]=0.0;
		 else break;
	} 

	if(k>0)
	{
		fprintf(cla->resultfp,"\n粗差总数：%d \n粗差估值：",k);
		for(int i=0;i<=cla->m_Lnumber-1;i++)
		{
			if(cla->P[i]>1.0e-10) continue;
			fprintf(cla->resultfp,"\n  %2d %8s %8s  %12.4lf ",
				i+1,cla->Pname[cla->StartP[i]],cla->Pname[cla->EndP[i]],cla->V[i]);
		}	
	}
	else fprintf(cla->resultfp,"\n粗差探测未发现粗差\n");
	
	cla->m_mu= sqrt(pvv/(cla->m_Lnumber-k-(cla->m_Pnumber-cla->m_knPnumber)));
	fprintf(cla->resultfp,"\n       μ=±%lf",cla->m_mu);

	cla->PrintResult(cla);

}

