// modern.cpp: implementation of the modern class.
//
//////////////////////////////////////////////////////////////////////


#include "modern.h"
#include "public.h"
#include <math.h>
#include <malloc.h>

#include "classic.h"

//    ===== 第三章 近代平差 =====
//    ===== 第三章 近代平差 =====
//    ===== 第三章 近代平差 =====
//    ===== 第三章 近代平差 =====




//////////////////////////////////////////////////////////////////////////
// 参数加权平差函数
// int n - 观测值个数
// int t - 参数个数
// double A[] - 误差方程系数矩阵，数组长度为n×t
// double L[] - 观测值向量,数组长度为n
// double P[] - 观测值权矩阵,仅存下三角元素，数组长度为n×(n+1)/2
// double X[] - 参数向量,函数调用前为参数的先验值,调用后为参数平差值，
//              数组长度为t
// double PX[] - 函数调用前为参数的先验权矩阵，调用后为参数平差值的权逆阵，
//               仅存下三角矩阵，数组长度为t×(t+1)/2 
// double V[] - 残差向量，数组长度为n
// 返回值：验后单位权中误差
//////////////////////////////////////////////////////////////////////////
double Adjust_With_PX(int n,int t,int t1,bool correlative,
					  double A[], double L[], double P[],
					  double X1[],double PX1[],int Num[],
					  double X[], double Qx[] ,double V[])
{
	double *U=(double *)malloc(sizeof(double)*t); //法方程自由项向量
	double *N=Qx;
		
	//  法方程系数矩阵与自由项清零 
	for(int i=0;i<t*(t+1)/2;i++) N[i]=0.0;
	for(int j=0;j<t;j++) U[j]=0.0;
	
	//组成法方程
	if(correlative) //观测值相关 
	{
		for(int k=0;k<n;k++)
		{
			for(int s=0;s<n;s++)
			{
				for(int i=0;i<t;i++)
				{
					U[i]+=A[k*t+i]*P[ij(k,s)]*L[s];
					for(int j=0; j<=i; j++)
					{
						N[ij(i,j)]+=A[k*t+i]*P[ij(k,s)]*A[s*t+j];
					}
				}
			}
		}
	}
	else  //观测值独立，P为对角矩阵 
	{
		for(int k=0;k<n;k++)
		{
			for(int i=0;i<t;i++)
			{
				U[i]+=A[k*t+i]*P[k]*L[k];
				for(int j=0; j<=i; j++)
				{
					N[ij(i,j)]+=A[k*t+i]*P[k]*A[k*t+j];
				}
			}
		}
	}

	for(int i=0;i<t1;i++) //先验信息处理
	{
		int k1 = Num[i];
		double xi=0.0;
		for(int j=0;j<t1;j++)
		{
			int k2 = Num[j];
			xi+=PX1[ij(i,j)]*X1[j];
			if(k1<=k2) N[ij(k1,k2)]+=PX1[ij(i,j)];
		}
		U[k1]+=xi;
	}
	
	if( !inverse(N,t)) 
	{
		free(U);
		return -1.0;
	}

	for(int i=0; i<t; i++)//计算参数平差值
	{
		double xi=0.0;
		for(int j=0; j<t; j++)
			xi+=N[ij(i,j)]*U[j];
		X[i]=xi;
	}

	for(int i=0; i<n; i++) //计算残差
	{
		double vi=-L[i];
		for(int j=0; j<t; j++) vi+=A[i*t+j]*X[j];
		V[i]=vi;
	}
	
	double *dX1=U;
	for(int i=0;i<t1;i++) // 计算随机参数的改正数
	{
		int ki=Num[i];
		dX1[i]=X1[i]-X[ki];
	}
	
	double pvv = Calculate_q(PX1,dX1,t1);
	if(correlative)pvv+=Calculate_q(P,V,n);
	else
	{
		for(int i=0;i<n;i++) pvv+=P[i]*V[i]*V[i];		
	}

	free(U);
	return sqrt(pvv/(n+t1-t));
}



//////////////////////////////////////////////////////////////////////////
// Kalman滤波函数
// int n - 观测值个数
// int t - 参数个数
// double A[] - 误差方程系数矩阵，数组长度为n×t
// double L[] - 观测值向量,数组长度为n
// double P[] - 观测值权矩阵,仅存下三角元素，数组长度为n×(n+1)/2
// double last_X[] - 上历元的参数向量，数组长度为t
// double F[] - 状态转移矩阵，数组长度t×t
// double X[] - 本历元的参数向量平差值, 数组长度为t
// double QW[] - 系统噪声权逆阵
// double Qx[] - 参数平差值的权逆阵，仅存下三角矩阵，数组长度为t×(t+1)/2 
// 返回值：计算成功返回true，失败返回false
//////////////////////////////////////////////////////////////////////////
bool KalmanFilter(int n,int t,double A[],double L[],double P[],
				  double F[],double QW[],double X[],double QX[])
{
	
	//  计算状态向量预报值
	double *last_X=(double *)malloc(sizeof(double)*t);
	for(int j=0;j<t;j++)
	{
		last_X[j]=X[j];
	}
	
	for(int i=0; i<t; i++)
	{
		X[i]=0.0;
		for(int j=0; j<t; j++)
			X[i] += F[i*t+j]*last_X[j];
	}
	free(last_X);

	//  计算状态预报值的权逆阵
	int tt=t*(t+1)/2;
	double *last_QX=(double *)malloc(sizeof(double)*tt);
	for(int j=0;j<tt;j++) last_QX[j]=QX[j];
	
	for(int j=0;j<tt;j++) QX[j]=QW[j];
	Calculate_BQBT(F,last_QX,t,t,QX);
	free(last_QX);

	//计算预报残差
	double *V=(double *)malloc(sizeof(double)*n);  //预测残差
	for(int i=0; i<n; i++)
	{
		V[i]=-L[i];
		for(int j=0;j<t;j++)
			V[i]+=A[i*t+j]*X[j];
	}
	
	//计算预报残差的权逆阵
	double *QV=(double *)malloc(sizeof(double)*n*(n+1)/2); 
	for(int i=0; i<n*(n+1)/2; i++)QV[i]=P[i];

	if(!inverse(QV,n)) return false;//观测值权矩阵变成权逆阵
	
	Calculate_BQBT(A,QX,n,t,QV);
	if(!inverse(QV,n))	return false;

	//计算状态估值的权逆阵
	double *M=(double *)malloc(sizeof(double)*n*t); //临时数组
	for(int i=0; i<n; i++ )
	{
		for(int j=0; j<t; j++)
		{
			double mij=0.0;
			for(int k=0;k<t;k++)
				mij+=A[i*t+k]*QX[ij(k,j)];
			M[i*t+j]=mij;
		}
	}
	
	//增益矩阵计算
	double *K=(double *)malloc(sizeof(double)*n*t); //增益矩阵
	for(int i=0; i<t; i++)
	{
		for(int j=0; j<n; j++)
		{
			double Kij=0.0;
			for(int k=0; k<t; k++)
			{
				Kij+=M[k*t+i]*QV[ij(k,j)];
			}
			K[i*n+j]=Kij;
		}
	}
	
	// 参数平差值计算
	for(int i=0; i<t; i++)
	{
		double xi=0;
		for(int j=0;j<n;j++)
			xi-=K[i*n+j]*V[j];
		X[i]+=xi;
	}


	for(int i=0; i<t; i++)
	{
		for(int j=0; j<=i; j++)
		{
			for(int k=0; k<n; k++)
				QX[ij(i,j)]-=K[i*n+k]*M[k*t+j];
		}
	}
	
	free(M);
	free(V);
	free(QV);
	free(K);

	return true;
}



//////////////////////////////////////////////////////////////////////////
//          抗差估计(观测值独立)
// 参数：
// int n - 观测值个数
// int t - 参数个数
// double A[] - 误差方程系数矩阵，数组长度为n×t
// double L[] - 观测值向量,数组长度为n
// double P[] - 观测值权矩阵的对角线元素,数组长度为n
// double (*ff)(double) - 权因子函数指针, 该函数参数和返回值均为双精度
// double sigma - 单位权中误差,用于计算标准化残差
// double eps - 迭代收敛条件
// double X[] - 参数平差值向量，数组长度为t
// double QX[] - 参数平差值的权逆阵，仅存下三角元素，数组长度为t×(t+1)/2
// double V[] - 残差向量，数组长度为n
// 返回值：验后单位权中误差
//////////////////////////////////////////////////////////////////////////
double Robust( int n, int t,double A[],double L[],double P[],
			   int ff,double k0,double k1, double sigma, double eps,
			  double X[], double QX[], double V[],double W[])
{
	double *N=QX;
	//---------------------------------------
	//    计算残差的中误差MV
	for(int i=0;i<t*(t+1)/2;i++) N[i]=0;
	for(int k=0; k<n; k++)
	{
		for(int i=0;i<t;i++)
		{
			for(int j=0;j<=i;j++)
			{
				N[ij(i,j)]+=A[k*t+i]*A[k*t+j]*P[k];
			}
		}  
	} 
	if(!inverse(N,t))
	{
		return -1.0;
	}
	
	double *MV=(double *)malloc(sizeof(double)*n); // MV-残差的中误差
	for(int i=0;i<n;i++) 
	{
		double q=1.0/P[i]-Calculate_q(N,A+i*t,t);
		MV[i]=sigma*sqrt(q);
	}			

	double *U =(double *)malloc(sizeof(double)*t);      //法方程自由项

	for(int s=1; ; s++) // s：迭代次数
	{
		
        for(int i=0;i<n;i++) // 计算等价权因子
		{
			double vi=-L[i];
			for(int j=0;j<t;j++)
			{
				vi+=A[i*t+j]*X[j];
			}
			W[i] = Wi(ff,vi/MV[i],k0,k1); 
		}
		
		// 组成法方程
		for(int i=0;i<t;i++) U[i]=0;
		for(int i=0;i<t*(t+1)/2;i++) N[i]=0;
		
		for(int k=0;k<n;k++)
		{
			double pk=P[k]*W[k]; //等价权
			for(int i=0;i<t;i++)
			{
				U[i]+=A[k*t+i]*pk*L[k];
				for(int j=0;j<=i;j++)
					N[ij(i,j)]+=A[k*t+i]*A[k*t+j]*pk;
			}  
		} 
		
		if(!inverse(N,t)) 
		{
			free(U);
			free(MV);
			return -1.0;
		}
		
		//计算参数平差值
		double max_dX=0.0; //X的最大变化量
		for(int i=0; i<t; i++)
		{
			double xi=0.0;
			for(int j=0;j<t;j++)
				xi+=N[ij(i,j)]*U[j];
			double dx=fabs(X[i]-xi);
			if(dx > max_dX) max_dX = dx;
			X[i]=xi;
		}   
		
		if(max_dX<eps && s>1)
		{
			double pvv=0.0;
			int nn=0;
			for(int i=0;i<n;i++) //计算残差及单位权中误差
			{
				double vi=-L[i];
				for(int j=0;j<t;j++)
				{
					vi+=A[i*t+j]*X[j];
				}
				
				pvv+=vi*vi*P[i]*W[i];
				V[i]=vi;
				if(W[i] > 0.0) nn++;
			}
			
			free(U);
			free(MV);
			
			return sqrt(pvv/(nn-t));  //验后单位权中误差
		}
	}

}





//////////////////////////////////////////////////////////////////////////
//	相关 robust 估计
// 参数：
// int n - 观测值个数
// int t - 参数个数
// double A[] - 误差方程系数矩阵，数组长度为n×t
// double L[] - 观测值向量,数组长度为n
// double P[] - 观测值权矩阵,仅存下三角元素，数组长度为n×(n+1)/2
// char *ff - 等价权函数名称字符串地址
// double sigma - 单位权中误差,用于计算标准化残差
// double eps - 迭代收敛条件
// double X[] - 参数平差值向量，数组长度为t
// double N[] - 参数平差值的权逆阵，仅存下三角元素，数组长度为t×(t+1)/2
// double V[] - 残差向量，数组长度为n
// 返回值：验后单位权中误差
//////////////////////////////////////////////////////////////////////////
//	相关 robust 估计
double CoRobust( int n, int t,double A[],double L[],double P[],
		        int ff,double k0,double k1,double sigma,double eps,
				double X[], double QX[], double V[],double W[])
{
	double *N=QX;
	//---------------------------------------
	//    计算残差的中误差MV
	for(int i=0;i<t*(t+1)/2;i++) N[i]=0.0;

	for(int i=0; i<t; i++)
	{
		for(int k=0;k<n;k++)
		{
			for(int s=0; s<n; s++)
			{
				double pks = P[ij(k,s)];
				for(int j=0;j<=i;j++)
					N[ij(i,j)]+=A[k*t+i]*pks*A[s*t+j];
			}
		}
	}

	if(!inverse(N,t))
	{
		fprintf(stderr,"CoRobust函数出错：LS估计法方程系数求逆失败！\n");
		return -1.0;
	}
	
	double *MV=(double *)malloc(sizeof(double)*n*(n+1)/2); // MV-残差的中误差
	for(int i=0; i<n*(n+1)/2; i++) MV[i]=P[i];
	if(!inverse(MV,n))
	{
		fprintf(stderr,"CoRobust函数出错：观测值权矩阵求逆失败！\n");
		return -1.0;
	}

	for(int i=0;i<n;i++) 
	{
		double q=MV[ij(i,i)]-Calculate_q(N,A+i*t,t);
		MV[i]=sigma*sqrt(q)+1.0e-12;
	}			
			 
	double *U = (double *)malloc(sizeof(double)*t);      //法方程自由项
	
    double m0; //验后单位权中误差
    for(int times=1; ; times++)
	{
		for(int i=0;i<n;i++) // 计算等价权因子
		{
			double vi=-L[i];
			for(int j=0;j<t;j++)
			{
				vi+=A[i*t+j]*X[j];
			}
			W[i] = Wi(ff,vi/MV[i],k0,k1); 
		}
		
		for(int i=0; i<t; i++) U[i]=0;

		for(int i=0; i<t*(t+1)/2; i++) N[i]=0.0;          

		//组成法方程
		for(int i=0; i<t; i++)
		{
		  for(int k=0;k<n;k++)
		  {
			  for(int s=0; s<n; s++)
			  {
				  double pks = P[ij(k,s)]*sqrt(W[k]*W[s]);
				  
				  U[i]+=A[k*t+i]*pks*L[s];
				  for(int j=0;j<=i;j++)
					  N[ij(i,j)]+=A[k*t+i]*pks*A[s*t+j];
			  }
		  }
		}
	     
 	    if(!inverse(N,t)) //法方程求逆
		{
            fprintf(stderr,"CoRobust出错：抗差估计法方程系数矩阵降秩！\n");
			return -1.0;
		}

		double max=0.0;
		for(int i=0; i<t; i++) //计算 X
		{
			double xi=0.0;
			for(int j=0;j<t;j++) xi+=N[ij(i,j)]*U[j];
			double dx = fabs(xi-X[i]);
			if(dx > max) max=dx;
			X[i]=xi;
		} 
	     
	    for(int i=0; i<n; i++)//  计算V
		{ 
			V[i] = -L[i];
			for(int j=0;j<t;j++) V[i]+=A[i*t+j]*X[j];
		} 

		if(max<eps && times>1)
		{
			for(int i=0; i<n; i++)
			{
				 double vi=-L[i];
				 for(int j=0;j<t;j++)
				 {
					 vi+=A[i*t+j]*X[j];
				 }
				 V[i]=vi;
			} 
				 
			// [pvv]计算
			double pvv=0.0;
			int new_n=0;
			for(int i=0; i<n; i++)
			{
				if( W[i] > 0.0 ) new_n++;
				for(int j=0;j<n;j++)
				{ 
			        pvv+=V[i]*P[ij(i,j)]*sqrt(W[i]*W[j])*V[j];
				}  
			}
			m0=sqrt(pvv/(new_n-t));
			break;
		} 
	} 

	free(U); 
	free(MV); 
	return  m0;	  
}






//////////////////////////////////////////////////////////////////////////
//	粗差探测，观测值不相关
// 参数：
// int n - 观测值个数
// int t - 参数个数
// double A[] - 误差方程系数矩阵，数组长度为n×t
// double L[] - 观测值向量,数组长度为n
// double P[] - 观测值权矩阵,仅存下三角元素，数组长度为n×(n+1)/2
// double eps - 迭代收敛条件,标准化残差的最大允许值：k×μ
// double m0  - 单位权中误差
// double X[] - 参数向量,函数调用前为参数的先验值,调用后为参数平差值，
//              数组长度为t
// double N[] - 参数平差值的权逆阵，仅存下三角矩阵，数组长度为t×(t+1)/2 
// double V[] - 残差向量，数组长度为n
// bool W[] - 粗差标志数组，false表示观测值存在粗差,数组长度为n
// 返回值：验后单位权中误差
//////////////////////////////////////////////////////////////////////////
double DataSnooping(int n,int t,double A[],double L[],double P[],
					double eps,double m0,
					double X[],double N[], double V[],bool W[])
{
	double *U=(double *)malloc(sizeof(double)*t);  //法方程自由项数组
	for(int i=0; i<n; i++) W[i]=true; // 剔除标志数组

	while(1)
	{
		for(int i=0; i<t; i++) U[i]=0;
		for(int i=0; i<t*(t+1)/2; i++) N[i]=0.0;
		
		//  组成法方程
		for(int k=0;k<n;k++)
		{
			if(!W[k])continue;
			double pk=P[k];
			for(int i=0;i<t;i++)
			{
				U[i]+=A[k*t+i]*pk*L[k];
				for(int j=0;j<=i;j++)
					N[ij(i,j)]+=pk*A[k*t+i]*A[k*t+j];
			} 
		} 
		  
		if(inverse(N,t)==false) //法方程系数矩阵求逆
		{
			free(U);
			return -1.0;
		}
		
		//   计算 X
		for(int i=0;i<t;i++)
		{
			X[i]=0.0;
			for(int j=0;j<t;j++)
				X[i]+=N[ij(i,j)]*U[j];
		} 
		 
		double pvv=0.0;
		int nn=n;
		for(int i=0;i<n;i++)
		{
			V[i]=-L[i];
			for(int j=0;j<t;j++)
			{
				V[i]+=A[i*t+j]*X[j];
			}
			if(W[i]) pvv+=P[i]*V[i]*V[i];
			else     nn--;
		}
		double m=sqrt(pvv/(nn-t));
		
		//  计算单位化残差,并挑选最大值
		double max_v=0.0;
		int    max_i;
		for(int i=0;i<n;i++)
		{
			if(!W[i])continue;
			double qv=1.0/P[i]-Calculate_q(N,A+i*t,t); //v的权倒数
			double mi=m0*sqrt(qv);
			double vi=fabs(V[i]/mi); //单位化残差
			if(vi>max_v)
			{
				max_v=vi;
				max_i=i;
			}
		}

		if(max_v<eps)
		{
			free(U);
			return m;
		}
		W[max_i]=false;
	}
}





//////////////////////////////////////////////////////////////////////////
//		序贯平差程序 sequential adjustment, successive adjustment
// 参数：
// int n - 观测值个数
// int t - 参数个数
// double A[] - 误差方程系数矩阵，数组长度为n×t
// double L[] - 观测值向量,数组长度为n
// double P[] - 观测值权矩阵,仅存下三角元素，数组长度为n×(n+1)/2
// double X[] - 参数向量,函数调用前为参数的先验值,调用后为参数平差值，
//              数组长度为t
// double Qx[] - 函数调用前为参数的先验权逆阵，调用后为参数平差值的权逆阵，
//               仅存下三角矩阵，数组长度为t×(t+1)/2 
// double V[] - 残差向量，数组长度为n
// 返回值：验后单位权中误差
//////////////////////////////////////////////////////////////////////////
double Successive_Adjustment( int n, int t,double A[],double L[],double P[],
				double X[], double QX[],double V[])
{
	// 预测残差计算
	double *VV=(double *)malloc(sizeof(double)*n);
	for(int i=0; i<n; i++)
	{
		double vi=-L[i];
		for(int k=0; k<t; k++)
			vi+=A[i*t+k]*X[k];
		VV[i]=vi;
	}
	
	double *Qv=(double *)malloc(sizeof(double)*n*(n+1)/2); //预测残差的权矩阵
	for(int i=0; i<n*(n+1)/2; i++)Qv[i]=P[i];
	if(!inverse(Qv,n))
	{
		if(VV) free(VV); //原文没有此处的两个free
		if(Qv) free(Qv);
		return -1.0;
	}

	Calculate_BQBT(A,QX,n,t,Qv);
	if(!inverse(Qv,n))
	{
		if(VV) free(VV); //原文没有此处的两个free
		if(Qv) free(Qv);
		return -1.0;
	}

	double *M=(double *)malloc(sizeof(double)*n*t); //临时数组，存放A*QX
	for(int i=0; i<n; i++)
	{
		for(int j=0; j<t; j++)
		{
			double mij=0.0;
			for(int k=0; k<t; k++)
				mij+=A[i*t+k]*QX[ij(k,j)];
			M[i*t+j]=mij;
		}
	}
	
	//计算增益矩阵 J=M'*PV
	double *J=(double *)malloc(sizeof(double)*n*t); //增益矩阵数组
	for(int i=0; i<t; i++)
	{
		for(int j=0; j<n; j++)
		{
			double Jij=0.0;
			for(int k=0; k<n; k++)
				Jij+=M[k*t+i]*Qv[ij(k,j)];
			J[i*n+j]=Jij;	
		}
	}
	
	// 计算X
	for(int i=0; i<t; i++)
	{
		double xi=0.0;
		for(int k=0; k<n; k++)
			xi-=J[i*n+k]*VV[k];
		X[i]=X[i]+xi;
	}

	// 计算QX
	for(int i=0; i<t; i++)
	{
		for(int j=0;j<=i;j++)
		{
			double qxij=0.0;
			for(int k=0; k<n; k++)
				qxij+=J[i*n+k]*M[k*t+j];
			QX[i*(i+1)/2+j]-=qxij;
		}
	}

	//残差计算
	for(int i=0; i<n; i++)
	{
		double vi=-L[i];
		for(int k=0; k<t; k++)
			vi+=A[i*t+k]*X[k];
		V[i]=vi;
	}
	
	double pvv=0.0;
	for(int i=0; i<n; i++)
	{
		for(int j=0;j<n;j++)
		{
			pvv+=V[j]*VV[i]*P[ij(i,j)];
		}
	}
	
	free(VV);
	free(Qv);
	free(M);
	free(J);
	
	return sqrt(pvv/n);
}






//////////////////////////////////////////////////////////////////////////
//   半参数估计
bool  SemiParameter(int n,int t,double alpha,
					double A[],double P[],double L[],double R[],
					double X[],double QX[],double V[])
{
	int nt=n+t; //法方程的阶数
	double *ATPL=(double *)malloc(sizeof(double)*nt); //法方程自由项数组
	
	//  法方程清零
	double *ATPA=QX;
	for(int i=0;i<nt*(nt+1)/2;i++) ATPA[i]=0.0;
	for(int j=0;j<nt;j++) ATPL[j]=0.0;
	
	//  组成法方程
	for(int k=0;k<n;k++)
	{
		for(int s=0;s<n;s++)
		{
			for(int i=0;i<t;i++)
			{
				ATPL[i]+=A[k*t+i]*P[ij(k,s)]*L[s];
				for(int j=0; j<=i; j++)
				{
					ATPA[ij(i,j)]+=A[k*t+i]*P[ij(k,s)]*A[s*t+j];
				}
			}
		}
	}
	
	// 计算PA
	for(int i=0; i<n; i++)
	{
		for(int j=0;j<t;j++)
		{
			double tmp=0.0;
			for(int k=0;k<n;k++)
			{
				tmp+=P[ij(i,k)]*A[k*t+j];
			}
			ATPA[ij(i+t,j)]=tmp;
		}
	}

	//   将P+α*R放在法矩阵的下三角矩阵中
	for(int i=0;i<n;i++)
	{
		for(int j=0;j<=i;j++)
		{
			ATPA[ij(i+t,j+t)]=P[ij(i,j)]+alpha*R[ij(i,j)];
		}
	}
	
	//---------------------------------------------
	//  计算PL
	for(int i=0; i<n; i++)
	{
		ATPL[i+t]=0.0;;
		for(int j=0;j<n;j++)
		{
			ATPL[i+t] += P[ij(i,j)]*L[j];
		} 
	}
	
	if(!inverse(ATPA,nt))
	{
		fprintf(stderr,"半参数估计出错：法方程系数矩阵求逆失败！");
		free(ATPL);
		return false;
	}
	
	//   参数、半参数计算
	for(int i=0;i<nt;i++)
	{
		X[i] = 0.0;
		for(int j=0;j<nt; j++)
		{
			X[i]+=ATPA[ij(i,j)]*ATPL[j];
		}
	}
	
	//  计算残差V
	for(int i=0;i<n;i++)
	{
		V[i] = -L[i]+X[i+t]; //非参数分量减去观测值
		for(int j=0;j<t; j++)
		{
			V[i]+=A[i*t+j]*X[j];
		}
	}

	return true;
	free(ATPL);
}


//////////////////////////////////////////////////////////////////////////
// 岭估计
// 返回值：验后单位权中误差
double Ridge(int n, int t,double A[],double L[],double P[],double k,
			 double X[],double QX[],double V[])
{
	double *U=(double *)malloc(sizeof(double)*t); //法方程自由项向量
	double *N=(double *)malloc(sizeof(double)*t*(t+1)/2);  //未加岭参数的系数矩阵
	double *NN=(double *)malloc(sizeof(double)*t*(t+1)/2); //加过岭参数的系数矩阵	
	
	//法方程系数阵和自由项清零
	for(int i=0; i<t; i++)
	{
		U[i]=0.0;
		for(int j=0;j<=i;j++) N[ij(i,j)]=0.0;
	}

	//  组成法方程式 
	for(int h=0;h<n;h++)
	{
		for(int s=0;s<n;s++)
		{
			for(int i=0;i<t;i++)
			{
				U[i]+=A[h*t+i]*P[ij(h,s)]*L[s];
				for(int j=0; j<=i; j++)
				{
					N[ij(i,j)]+=A[h*t+i]*P[ij(h,s)]*A[s*t+j];
				}
			}
		}
	}

	//添加岭参数
	for(int i=0; i<t; i++)
	{
		for(int j=0;j<i;j++)
			NN[ij(i,j)]=N[ij(i,j)];
		NN[ij(i,i)]=N[ij(i,i)]+k;
	}
	
	//法方程系数矩阵求逆
	if(!inverse(NN,t))
	{
		free(U);
		free(N);
		free(NN);
		return -1.0;
	}
	
	//计算X
	for(int i=0;i<t;i++)
	{
		X[i]=0.0;
		for(int j=0;j<t;j++)X[i]+=NN[ij(i,j)]*U[j];
	}
	free(U);
	
	//  计算QX 
	for(int h=0; h<t; h++)
	{
		for(int s=0;s<=h;s++)
		{
			double q=0.0;
			for(int j=0;j<t;j++)
			{
				for(int i=0; i<t; i++)
				{
					q+=NN[ij(h,j)]*N[ij(j,i)]*NN[ij(i,s)];
				}
			}  	
			QX[ij(h,s)]=q;
		}
	}
	free(N);
	free(NN);

	//  计算V
	for(int i=0;i<n;i++)
	{
		double vi=-L[i];
		for(int j=0;j<t;j++) vi+=A[i*t+j]*X[j];
		V[i]=vi;
	}

	if(n==t)return 0.0;
	double pvv=Calculate_q(P,V,n);
	
	return sqrt(pvv/(n-t));
}


