// MyGPS.cpp: implementation of the CMyGPS class.
//
//////////////////////////////////////////////////////////////////////


#include "mygps.h"
#include "public.h"
#include <malloc.h>

#define Old_Method  1
#define New_Method  0
#define Method  Old_Method


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

void CCoGPS_CCoGPS(struct CCoGPS *ccgps)
{
	ccgps->m_IsPx = false; //不进行参数加权平差
	ccgps->XYZ=NULL;     // 坐标数组	
    	ccgps->Pname=NULL;  // 点名指针数组
	ccgps->IsKnown=NULL;//已知点标志数组
	
    	ccgps->dir0=NULL;        //各同步区首向量在观测值中的序号
    	ccgps->dir1=NULL;
    	ccgps->dir2=NULL;  //向量端点号

	ccgps->L=NULL;  //全区观测向量
	ccgps->ATPA=NULL;
	ccgps->ATPL=NULL;
	ccgps->P=NULL;     //权矩阵数组；
	ccgps->QV=NULL;     //观测值残差权倒数
	ccgps->V=NULL;      //残差
    	ccgps->PV=NULL;     //
	ccgps->W=NULL;      //权因子	

    	ccgps->MxMyMz=NULL;  //已知点的中误差向量
	ccgps->KnownXYZ=NULL;  //已知点坐标,参数加权平差时用
	ccgps->KnownPoints=NULL; //与已知点坐标相应的点号
    	ccgps->KnownPx=NULL;  //已知点权矩阵
	ccgps->AreaNumber=NULL;
	
	
	ccgps->CCoGPS=CCoGPS_CCoGPS;
	ccgps->de_CCoGPS=CCoGPS_de_CCoGPS;
	ccgps->InputData=CCoGPS_InputData;
	ccgps->InputKnownXYZ=CCoGPS_InputKnownXYZ;
	ccgps->PrintData=CCoGPS_PrintData;
	ccgps->PrintXYZ=CCoGPS_PrintXYZ;
	ccgps->PrintLV=CCoGPS_PrintLV;
	ccgps->LeastSquares=CCoGPS_LeastSquares;
	ccgps->Free=CCoGPS_Free;
	ccgps->Quasi_Stable=CCoGPS_Quasi_Stable;
	ccgps->Robust=CCoGPS_Robust;
	ccgps->DataSnooping=CCoGPS_DataSnooping;
	ccgps->Helmert=CCoGPS_Helmert;
	ccgps->CaATPA=CCoGPS_CaATPA;
	ccgps->CaATPAi=CCoGPS_CaATPAi;
	ccgps->Known=CCoGPS_Known;
	ccgps->CaLV=CCoGPS_CaLV;
	ccgps->CaVTPV=CCoGPS_CaVTPV;
	ccgps->Ca_dX=CCoGPS_Ca_dX;
	ccgps->GetStationNumber=CCoGPS_GetStationNumber;
	ccgps->Ca_VTPV_XYZ=CCoGPS_Ca_VTPV_XYZ;
	ccgps->CaJTPJ=CCoGPS_CaJTPJ;
	ccgps->CaQg=CCoGPS_CaQg;
	ccgps->CaJTPA=CCoGPS_CaJTPA;
	ccgps->Ca_X_Qx=CCoGPS_Ca_X_Qx;
	ccgps->Ca_UintV=CCoGPS_Ca_UintV;
	ccgps->caQV=CCoGPS_caQV;
}


void CCoGPS_de_CCoGPS(struct CCoGPS *ccgps)
{
  if(ccgps->XYZ) free(ccgps->XYZ);
  if(ccgps->Pname) free(ccgps->Pname);
  if(ccgps->IsKnown) free(ccgps->IsKnown);
  if(ccgps->dir0) free(ccgps->dir0);
  if(ccgps->dir1) free(ccgps->dir1);
  if(ccgps->dir2) free(ccgps->dir2);
  if(ccgps->L) free(ccgps->L);
  if(ccgps->ATPA) free(ccgps->ATPA);
  if(ccgps->ATPL) free(ccgps->ATPL);
  if(ccgps->P) free(ccgps->P);
  if(ccgps->QV) free(ccgps->QV);
  if(ccgps->V) free(ccgps->V);
  if(ccgps->PV) free(ccgps->PV);
  if(ccgps->W) free(ccgps->W);
  if(ccgps->MxMyMz) free(ccgps->MxMyMz);
  if(ccgps->KnownXYZ) free(ccgps->KnownXYZ);
  if(ccgps->KnownPoints) free(ccgps->KnownPoints);
  if(ccgps->KnownPx) free(ccgps->KnownPx);
  if(ccgps->AreaNumber) free(ccgps->AreaNumber);
}



/////////////////////////////////////////////////////////////
//        一个同步区的法方程,用于最小二乘平差
void CCoGPS_CaATPAi(int nk,  //观测值总数，不是向量总数
					 int *dir1,
					 int *dir2,
					 double *Pk,
					 double *Lk ,// 自由项
					 struct CCoGPS *ccgps
					)  
{
     for(int s1=0;s1<nk;s1++)
     {
         int i1=3*dir1[s1/3]+s1%3; //误差方程Vk中系数为-1的未知数编号
         int i2=3*dir2[s1/3]+s1%3; //误差方程Vk中系数为+1的未知数编号

         for(int s2=0; s2<nk; s2++)
         {
            	int j1=3*dir1[s2/3]+s2%3;
            	int j2=3*dir2[s2/3]+s2%3;

	    	double p12 = Pk[ij(s1,s2)];

            	double l=Lk[s2];

            	ccgps->ATPL[i1]-=p12*l;
            	ccgps->ATPL[i2]+=p12*l;

            	if(i1>=j1)ccgps->ATPA[ij(i1,j1)]+=p12;
            	if(i1>=j2)ccgps->ATPA[ij(i1,j2)]-=p12;
            	if(i2>=j1)ccgps->ATPA[ij(i2,j1)]-=p12;
            	if(i2>=j2)ccgps->ATPA[ij(i2,j2)]+=p12;
         }
     }
	 
}




//////////////////////////////////////////////////////////////////////////
//    组成法方程式
void CCoGPS_CaATPA(struct CCoGPS *ccgps)
{
     int t=3*ccgps->Pnumber;
     int tt = t*(t+1)/2;
     for(int i=0;i<t;i++) ccgps->ATPL[i]=0.0;
     for(int i=0;i<tt;i++) ccgps->ATPA[i]=0.0;

     double *l=(double *)malloc(sizeof(double)*3*ccgps->Vnumber);//new double [3*ccgps->Vnumber];//误差方程自由项
	 ccgps->CaLV(l,ccgps); //误差方程自由项计算

     for(int i=0;i<ccgps->Anumber;i++) //逐向量组组成法方程
     {
         int s=ccgps->dir0[i];
         int n=ccgps->dir0[i+1]-s;
         ccgps->CaATPAi(3*n,ccgps->dir1+s,ccgps->dir2+s,ccgps->P[i],l+3*s,ccgps);
     }

     free(l);
}






//////////////////////////////////////////////////////////////////////
//      输入原始数据（近似坐标、GPS向量）
bool CCoGPS_InputData(char *file,struct CCoGPS *ccgps)
{
	 FILE *fp=fopen(file,"r"); // 打开原始数据文件
	 if(fp==NULL)
	 {
		 fprintf(stderr,"原始数据文件打不开!\n");
		 return 0;
	 }

	 fscanf(fp,"%d",&ccgps->Pnumber); //总点数
	 fscanf(fp,"%d",&ccgps->Anumber); //同步区总数（向量组总数）
	 fscanf(fp,"%d",&ccgps->Vnumber); //向量总数
	 
	 ccgps->Pname = (char **)malloc(sizeof(char *)*ccgps->Pnumber);//new char* [ccgps->Pnumber]; //点名地址数据
	 ccgps->IsKnown = (bool *)malloc(sizeof(bool)*ccgps->Pnumber);//new bool[ccgps->Pnumber]; //已知/未知点标志数组
	 for(int i=0;i<ccgps->Pnumber;i++)ccgps->Pname[i]=NULL;
	 
     	ccgps->dir0=(int *)malloc(sizeof(int)*(ccgps->Vnumber+1));//new int [ccgps->Anumber+1];//各同步区首向量在总观测向量中的序号
     	ccgps->dir1=(int *)malloc(sizeof(int)*ccgps->Vnumber);//new int [ccgps->Vnumber];  //向量起始点号
     	ccgps->dir2=(int *)malloc(sizeof(int)*ccgps->Vnumber);//new int [ccgps->Vnumber];  //向量终点点号
	 ccgps->AreaNumber=(int *)malloc(sizeof(int)*ccgps->Vnumber);//new int [ccgps->Vnumber]; //各向量所属向量组的编号
	 
     int n=3*ccgps->Vnumber;  //观测值总数
     ccgps->L=(double *)malloc(sizeof(double)*n);//new double [n]; //向量观测值
     ccgps->V=(double *)malloc(sizeof(double)*n);//new double [n]; //残差
 
	 ccgps->P=(double **)malloc(sizeof(double *)*ccgps->Anumber);//new double* [ccgps->Anumber]; //观测值权矩阵地址
     ccgps->PV=(double *)malloc(sizeof(double)*n);//new double [n]; //权矩阵于残差向量的乘积
	 
     int t=3*ccgps->Pnumber;
     int tt=t*(t+1)/2;
	 ccgps->XYZ =(double *)malloc(sizeof(double)*t);//new double[t];   //坐标数组
     ccgps->ATPA=(double *)malloc(sizeof(double)*tt);//new double [tt]; //法方程系数矩阵（或逆矩阵）
     ccgps->ATPL=(double *)malloc(sizeof(double)*t);//new double [t];  //法方程自由项向量

	 ccgps->unPnumber=0; //未知点总数
	 for(int k=0;k<ccgps->Pnumber;k++) //输入坐标
	 {
		 char name[20];
		 fscanf(fp,"%s",name); //点名
		 int i=ccgps->GetStationNumber(name,ccgps); //将点名换算成点号
		 if(i<0)
		 {
			 fprintf(stderr,"读取坐标是出错：点名多于总点数\n");  
			 return 0;
		 }
		 
		 fscanf(fp,"%lf%lf%lf",ccgps->XYZ+3*i,ccgps->XYZ+3*i+1,ccgps->XYZ+3*i+2);//坐标值 
		 int ppppppIsKnown;
		 fscanf(fp,"%d",&ppppppIsKnown); //已知点标志   ////////////////////////////////////////////////////////////?????
		 ppppppIsKnown?((*(ccgps->IsKnown+i))=true):((*(ccgps->IsKnown+i))=false);
		 if(!ccgps->IsKnown[i])ccgps->unPnumber++;
	 }

     ccgps->dir0[0]=0;
     for(int i=0;i<=ccgps->Anumber-1;i++)   //逐同步区循环
     {
		 int d;
		 fscanf(fp,"%d",&d); //同步区编号		 
		 if(d!=i+1)
		 {
			 fprintf(stderr,"同步区编号不连续\n");
			 return 0;
		 }

         int jd;  //同步区内向量数
		 fscanf(fp,"%d",&jd);

		 ccgps->dir0[i+1]=ccgps->dir0[i]+jd;
         for(int j=ccgps->dir0[i];j<ccgps->dir0[i+1];j++)
         {			 
             char P1[20],P2[20];
			 fscanf(fp,"%s %s",P1,P2); //输入向量的端点点名
             ccgps->dir1[j]=ccgps->GetStationNumber(P1,ccgps);
             ccgps->dir2[j]=ccgps->GetStationNumber(P2,ccgps);
			 if(ccgps->dir1[j]<0 || ccgps->dir2[j]<0)
			 {
				 fprintf(stderr,"读取向量数据出错：点名多余总点数\n");
				 return 0;
			 }

             //输入向量观测值
			 fscanf(fp,"%lf %lf %lf",ccgps->L+3*j,ccgps->L+3*j+1,ccgps->L+3*j+2);
			 ccgps->AreaNumber[j]=i;
         }

		 //---------------------------------------
		 //     输入方差协方差矩阵
		 int ni=3*jd;
         	double *Pi=(double *)malloc(sizeof(double)*(ni+1)*ni/2);//new double [(ni+1)*ni/2];
		 ccgps->P[i]=Pi;		 
		 
	 	 int s=0;
         	for(int j=0;j<ni;j++)
		 {
			 int d;
			 fscanf(fp,"%d",&d); //行号
		     for(int k=0;k<=j;k++)
			 {
				 fscanf(fp,"%lf",Pi+s); //向量的方差协方差
				 s++;
			 }
		 }
     }
     fclose(fp);

	 return 1;
}


//////////////////////////////////////////////////////////////////////
//       输出原始数据
void CCoGPS_PrintData(struct CCoGPS *ccgps)
{
	fprintf(ccgps->resultfp,"\n  总点数: %d",ccgps->Pnumber);
	fprintf(ccgps->resultfp,"\n  向量组总数: %d",ccgps->Anumber);
	fprintf(ccgps->resultfp,"\n  向量总数: %d\n",ccgps->Vnumber);

	//输出已知坐标和近似坐标	
	for(int k=0; k<ccgps->Pnumber; k++) 
	{
		fprintf(ccgps->resultfp,"\n %4s",ccgps->Pname[k]); //输出点名
		fprintf(ccgps->resultfp,"  %14.5lf",ccgps->XYZ[3*k]);
		fprintf(ccgps->resultfp,"  %14.5lf",ccgps->XYZ[3*k+1]);
		fprintf(ccgps->resultfp,"  %14.5lf",ccgps->XYZ[3*k+2]);
		fprintf(ccgps->resultfp,"  %d",ccgps->IsKnown[k]);
	}		
	 
	//输出向量观测值及其方差阵
	for(int i=0; i<ccgps->Anumber; i++)
	{
		 int jd=ccgps->dir0[i+1]-ccgps->dir0[i]; //组内向量总数
         fprintf(ccgps->resultfp,"\n\n%4d  %d",i+1,jd);//向量组编号，组内向量数

         for(int j=ccgps->dir0[i];j<ccgps->dir0[i+1];j++)
         {
             int k1=ccgps->dir1[j];
             int k2=ccgps->dir2[j];

             double dx=ccgps->L[3*j];
             double dy=ccgps->L[3*j+1];
             double dz=ccgps->L[3*j+2];
	         
	     fprintf(ccgps->resultfp,"\n %8s %8s",ccgps->Pname[k1],ccgps->Pname[k2]);
	     fprintf(ccgps->resultfp,"%17.5lf %17.5lf %17.5lf ",dx,dy,dz);
	 }

         int jj=0;
         for(int j=0;j<3*jd;j++)
         {
	      fprintf(ccgps->resultfp,"\n%3d",j+1);
              for(int k=0;k<=j;k++)
			  {
				   if(k%4==0 && k>0)fprintf(ccgps->resultfp,"\n   "); 
				   fprintf(ccgps->resultfp," %17.9e",ccgps->P[i][jj++]);
			  }
		 }
	 }
}


//////////////////////////////////////////////////////////////////////////
//    未知数计算
double CCoGPS_Ca_dX(bool IsPrint,struct CCoGPS *ccgps)
{
     int t=3*ccgps->Pnumber;
     if(!inverse(ccgps->ATPA,t))
	 {
		 fprintf(stderr,"法方程不满秩!\n");
		 fclose(ccgps->resultfp);
		 exit(0);
	 }
     
	 if(IsPrint)
	 {
		 fprintf(ccgps->resultfp,"\n\n           ===== 坐标改正数 =====\n");		 
	 }
	 double max=0.0;
	 for(int i=0;i<t;i++)
	 {
          double xi=0.0;
   	      for(int j=0;j<t;j++)xi-=ccgps->ATPA[ij(i,j)]*ccgps->ATPL[j];
		  ccgps->XYZ[i]+=xi;
		  if(fabs(xi)>fabs(max))max=xi;
		  if(IsPrint)
		  {
			  if(i%3==0) fprintf(ccgps->resultfp,"\n%10s ",ccgps->Pname[i/3]);
			  fprintf(ccgps->resultfp," %8.4lf",xi);
		  }
     }
	 return max;
}


//////////////////////////////////////////////////////////////////////
//   计算误差方程自由项或最小二乘残差
void CCoGPS_CaLV(double V[],struct CCoGPS *ccgps)
{
     for(int i=0;i<ccgps->Vnumber;i++)
     {
         int k1=ccgps->dir1[i];
         int k2=ccgps->dir2[i];
         
	 ccgps->V[3*i]=ccgps->XYZ[3*k2]-ccgps->XYZ[3*k1]-ccgps->L[3*i];         
         ccgps->V[3*i+1]=ccgps->XYZ[3*k2+1]-ccgps->XYZ[3*k1+1]-ccgps->L[3*i+1]; 
         ccgps->V[3*i+2]=ccgps->XYZ[3*k2+2]-ccgps->XYZ[3*k1+2]-ccgps->L[3*i+2]; 
     }
}

//////////////////////////////////////////////////////////////////////
//    计算 PV,VTPV
double CCoGPS_CaVTPV(struct CCoGPS *ccgps)
{
	 double VTPV=0.0;
	 for(int i=0;i<ccgps->Anumber;i++) // Anumber：向量组总数
	 {
	    	double *Pi=ccgps->P[i];     //i区权矩阵起始地址
        	int k1=3*ccgps->dir0[i];    //i区首向量在观测值中的序号
        	int k2=3*ccgps->dir0[i+1];  //i+1区首向量在观测值中的序号

		double *PVi=ccgps->PV+k1;
		double *Vi=ccgps->V+k1;
		int n=k2-k1;

		double pvv=0.0;
		for(int j=0;j<n;j++)
		{
			double pvj=0.0;
			for(int k=0; k<n; k++) pvj+=Pi[ij(j,k)]*Vi[k];
			PVi[j]=pvj;
			pvv += pvj*Vi[j];
		}
		VTPV+=pvv;
	 }
	 return VTPV;
}



//////////////////////////////////////////////////////////////////////
//    输出残差V,也可以输出自由项向量，还可以输出权因子
void  CCoGPS_PrintLV(char *Title,double *V,struct CCoGPS *ccgps)
{
	fprintf(ccgps->resultfp,"\n\n        	         ===== %s =====\n\n",Title);
	fprintf(ccgps->resultfp,"    向量组  点 名       点 名"
		"     Vx         Vy         Vz");

	int last_ni = -1;
	for(int i=0;i<ccgps->Vnumber;i++)
	{
		int ni=ccgps->AreaNumber[i];
		if(ni!=last_ni)	fprintf(ccgps->resultfp,"\n");
		last_ni=ni; 
		
		fprintf(ccgps->resultfp,"\n%8d",ni+1);
		fprintf(ccgps->resultfp,"%10s %10s",ccgps->Pname[ccgps->dir1[i]],ccgps->Pname[ccgps->dir2[i]]);
		fprintf(ccgps->resultfp,"%10.4lf %10.4lf %10.4lf",ccgps->V[3*i],ccgps->V[3*i+1],ccgps->V[3*i+2]);
	}  
}




//////////////////////////////////////////////////////////////////////
//  已知点处理（已知点固定，已知点加权）
void CCoGPS_Known(struct CCoGPS *ccgps)
{
	if(!ccgps->m_IsPx) //已知坐标固定平差
	{
		for(int i=0;i<ccgps->Pnumber;i++)
		{
			if(ccgps->IsKnown[i])
			{
				ccgps->ATPA[ij(3*i,3*i)]+=1.0e30;
				ccgps->ATPA[ij(3*i+1,3*i+1)]+=1.0e30;
				ccgps->ATPA[ij(3*i+2,3*i+2)]+=1.0e30;
			}
		}
		return;
	}

    //--------------------------------------
	//  已知坐标加权平差
	int kn=ccgps->Pnumber-ccgps->unPnumber;
	for(int i=0;i<3*kn;i++) //系数矩阵
	{
		int p1=3*ccgps->KnownPoints[i/3]+i%3;
		for(int j=0;j<=i;j++)
		{
			int p2=3*ccgps->KnownPoints[j/3]+j%3;
			ccgps->ATPA[ij(p1,p2)]+=ccgps->KnownPx[ij(i,j)];
		} 
	} 
		 
	//法方程自由项处理
	double *dX=(double *)malloc(sizeof(double)*3*kn);//new double[3*kn]; //已知坐标与近似坐标的坐标差
	for(int i=0;i<kn;i++)
	{
		int k=ccgps->KnownPoints[i];
		dX[3*i]  =ccgps->XYZ[3*k]  -ccgps->KnownXYZ[3*i];
		dX[3*i+1]=ccgps->XYZ[3*k+1]-ccgps->KnownXYZ[3*i+1];
		dX[3*i+2]=ccgps->XYZ[3*k+2]-ccgps->KnownXYZ[3*i+2];
	}

	for(int i=0;i<3*kn;i++)
	{
		double dL=0.0;
		for(int j=0;j<3*kn;j++)
			dL+=ccgps->KnownPx[ij(i,j)]*dX[j];
		int k=ccgps->KnownPoints[i/3]; //点号
		ccgps->ATPL[3*k+i%3]+=dL;
	} 
	free(dX);
	return;
}



//////////////////////////////////////////////////////////////////////
//   自由网平差函数
void CCoGPS_Free(struct CCoGPS *ccgps)
{
	//将观测值的方差协方差矩阵求逆，化为权矩阵
	for(int i=0; i<ccgps->Anumber; i++)
	{
		int ni = 3*(ccgps->dir0[i+1]-ccgps->dir0[i]); //向量组观测值总数
		inverse(ccgps->P[i],ni);
	}
	
	ccgps->CaATPA(ccgps);     	 
	for(int i=0; i<3*ccgps->Pnumber; i++)
	{
		for(int j=0; j<=i;j++)
		{
			if(i%3==j%3) ccgps->ATPA[ij(i,j)]+=1.0/ccgps->Pnumber;
		}
	}
	
	double max=ccgps->Ca_dX(true,ccgps);     
	for(int i=0; i<3*ccgps->Pnumber; i++)
	{
		for(int j=0; j<=i;j++)
		{
			if(i%3==j%3) ccgps->ATPA[ij(i,j)]-=1.0/ccgps->Pnumber;
		}
	}
	
	ccgps->CaLV(ccgps->V,ccgps);  //计算残差
	
	double VTPV=ccgps->CaVTPV(ccgps);
	ccgps->Sigma=sqrt(VTPV/(3*ccgps->Vnumber-3*(ccgps->Pnumber-1)));

	ccgps->PrintXYZ(ccgps);
	ccgps->PrintLV("V",ccgps->V,ccgps);
	fprintf(ccgps->resultfp,"\n\n自由网平差:\n[pvv]=%e",VTPV);
	fprintf(ccgps->resultfp,"\n单位权中误差μ=±%e",ccgps->Sigma);	
	
}




//////////////////////////////////////////////////////////////////////
//  最小二乘平差函数
void CCoGPS_LeastSquares(struct CCoGPS *ccgps)
{
	//将观测值的方差协方差矩阵求逆，化为权矩阵
	for(int i=0; i<ccgps->Anumber; i++)
	{
		int ni = 3*(ccgps->dir0[i+1]-ccgps->dir0[i]); //向量组观测值总数
		inverse(ccgps->P[i],ni);
	}
	
	ccgps->CaATPA(ccgps);     	 
	ccgps->Known(ccgps);
	double max=ccgps->Ca_dX(true,ccgps);     
	
	ccgps->CaLV(ccgps->V,ccgps);  //计算最小二乘残差
	double VTPV=ccgps->CaVTPV(ccgps);

	if(ccgps->m_IsPx)
	{
		double pvv=ccgps->Ca_VTPV_XYZ(ccgps)+VTPV;
		ccgps->Sigma=sqrt(pvv/(3*ccgps->Vnumber-3*ccgps->unPnumber));
	}
	else
	{
		ccgps->Sigma=sqrt(VTPV/(3*ccgps->Vnumber-3*ccgps->unPnumber));
	}	
	ccgps->PrintXYZ(ccgps);
	ccgps->PrintLV("V",ccgps->V,ccgps);

	fprintf(ccgps->resultfp,"\n\n最小二乘平差:\n[pvv]=%e",VTPV);
	fprintf(ccgps->resultfp,"\n单位权中误差μ=±%e",ccgps->Sigma);	
			 
}

//////////////////////////////////////////////////////////////////////
//  最小二乘平差函数
void CCoGPS_Helmert(struct CCoGPS *ccgps)
{

	for(int kk=1;kk<=2;kk++)
	{
		ccgps->CaATPA(ccgps);     	 
		ccgps->Known(ccgps);
		double max=ccgps->Ca_dX(1,ccgps);     
		
		ccgps->CaLV(ccgps->V,ccgps);  //计算最小二乘残差
		ccgps->PrintLV("V",ccgps->V,ccgps);
		
		double VTPV=0.0;
		
		for(int i=0;i<ccgps->Anumber;i++) // Anumber：向量组总数
		{
			double *Pi=ccgps->P[i];     //i区权矩阵起始地址
			int k1=3*ccgps->dir0[i];    //i区首向量在观测值中的序号
			int k2=3*ccgps->dir0[i+1];  //i+1区首向量在观测值中的序号
			
			double *PVi=ccgps->PV+k1;
			double *Vi=ccgps->V+k1;
			int n=k2-k1;
			
			double pvv=0.0;
			int d0 = 3*ccgps->dir0[i];
			for(int j=0;j<n;j++)
			{
				double pvj=0.0;
				for(int k=0; k<n; k++) pvj+=Pi[ij(j,k)]*Vi[k];
				
				PVi[j]=pvj;
				pvv += pvj*Vi[j];
			}
			
			
			fprintf(ccgps->resultfp,"\n[%d]=%lf  \n",i+1,pvv/n);

			double m2=pvv/n;
			//if(m2<0.1)m2=0.1;
			
			for(int j=0; j<(n+1)*n/2; j++)Pi[j]/=m2;
			
			//if(kk==2)inverse(Pi,n);
			VTPV+=pvv;
		}

			
	}
//	PrintData();
//	return;

	double VTPV=ccgps->CaVTPV(ccgps);
	if(ccgps->m_IsPx)
	{
		double pvv=ccgps->Ca_VTPV_XYZ(ccgps)+VTPV;
		ccgps->Sigma=sqrt(pvv/(3*ccgps->Vnumber-3*ccgps->unPnumber));
		
	}
	else
	{
		ccgps->Sigma=sqrt(VTPV/(3*ccgps->Vnumber-3*ccgps->unPnumber));
	}	
	ccgps->PrintXYZ(ccgps);
	ccgps->PrintLV("V",ccgps->V,ccgps);
	
	fprintf(ccgps->resultfp,"\n     验后精度估计:\n");
	fprintf(ccgps->resultfp,"\n     [pvv]:    %e",VTPV);
	fprintf(ccgps->resultfp,"\n     m:        %e(验后单位权中误差)",ccgps->Sigma);
}




//////////////////////////////////////////////////////////////////////
// 计算残差V的权倒数
void CCoGPS_caQV(struct CCoGPS *ccgps)
{
	//计算观测值平差值权倒数
	for(int i=0;i<3*ccgps->Vnumber; i++)
    	{
        int k1=3*ccgps->dir1[i/3]+i%3;  //测站点未知数编号
        int k2=3*ccgps->dir2[i/3]+i%3;  //照准点未知数编号
		
		double q=ccgps->ATPA[ij(k1,k1)]+ccgps->ATPA[ij(k2,k2)]-2.0*ccgps->ATPA[ij(k1,k2)];

		ccgps->QV[i] = ccgps->QV[i]-q;
    }	
}



//////////////////////////////////////////////////////////////////////
//  计算中位数、:   (1)计算单位权残差; (2)中位数计算方差因子
void CCoGPS_Ca_UintV(struct CCoGPS *ccgps)
{
	for(int i=0; i<3*ccgps->Vnumber; i++)
	{
		double mi=sqrt(ccgps->QV[i]);
		if(fabs(ccgps->QV[i])<1.0e-15) ccgps->W[i] = 1.0;
		else ccgps->W[i]=ccgps->V[i]/mi;  
	}

    ccgps->MEDIAN=Median(ccgps->W,3*ccgps->Vnumber,true)*1.4826;
	
	//fprintf(resultfp,"\nMEDIAN=%10e\n",MEDIAN);
}






//////////////////////////////////////////////////////////////////////////
//     输出坐标平差值及其中误差
void CCoGPS_PrintXYZ(struct CCoGPS *ccgps)
{
	double m=ccgps->Sigma;
	double Maxmx=0.0;
	double Maxmy=0.0;
	double Maxmz=0.0;

	fprintf(ccgps->resultfp,"\n\n\n STATION       X              Y              Z          RMS_X  RMS_Y  RMS_Z");
	fprintf(ccgps->resultfp,"\n              (m)            (m)            (m)          (cm)   (cm)   (cm)\n");
	fprintf(ccgps->resultfp," ----------------------------------------------------------------------------\n");

	for(int k=0;k<ccgps->Pnumber;k++)
	{
		fprintf(ccgps->resultfp,"\n%4d %4s",k+1,ccgps->Pname[k]);

		double X=ccgps->XYZ[3*k];
		double Y=ccgps->XYZ[3*k+1];
		double Z=ccgps->XYZ[3*k+2];				 

		fprintf(ccgps->resultfp," %14.4lf %14.4lf %14.4lf",X,Y,Z);

		double mx,my,mz;

		mx=sqrt(ccgps->ATPA[ij(3*k,3*k)])*m*100.0;
		my=sqrt(ccgps->ATPA[ij(3*k+1,3*k+1)])*m*100.0;
		mz=sqrt(ccgps->ATPA[ij(3*k+2,3*k+2)])*m*100.0;

		fprintf(ccgps->resultfp," %6.2lf %6.2lf %6.2lf",mx,my,mz);

		if(mx>Maxmx)Maxmx=mx;
		if(my>Maxmy)Maxmy=my;
		if(mz>Maxmz)Maxmz=mz;

	}
	fprintf(ccgps->resultfp,"\n\n  Max(RMS_X)=%5.2lf",Maxmx);
	fprintf(ccgps->resultfp,"\n  Max(RMS_Y)=%5.2lf",Maxmy);
	fprintf(ccgps->resultfp,"\n  Max(RMS_Z)=%5.2lf",Maxmz);
}




//////////////////////////////////////////////////////////////////////////
//    已知点加（相关）权平差时，坐标改正数的二次型
double CCoGPS_Ca_VTPV_XYZ(struct CCoGPS *ccgps)
{
	 int kn=ccgps->Pnumber-ccgps->unPnumber; //已知点数
	 double *Vxyz=(double *)malloc(sizeof(double)*3*kn);//new double [3*kn];

	 for(int i=0;i<kn;i++)
	 {
		 int k=ccgps->KnownPoints[i];  //已知点的点号
		 double X=ccgps->XYZ[3*k];
		 double Y=ccgps->XYZ[3*k+1];
		 double Z=ccgps->XYZ[3*k+2];				 

		 Vxyz[3*i]  =X-ccgps->KnownXYZ[3*i];				 
		 Vxyz[3*i+1]=Y-ccgps->KnownXYZ[3*i+1];				 
		 Vxyz[3*i+2]=Z-ccgps->KnownXYZ[3*i+2];
	 }

	 double VTPV_XYZ=0.0;
	 for(int i=0;i<3*kn;i++)
	 {
		 for(int j=0;j<3*kn;j++)
			 VTPV_XYZ+=Vxyz[i]*ccgps->KnownPx[ij(i,j)]*Vxyz[j];
	 }

	 free(Vxyz);
	 return VTPV_XYZ;
}




//////////////////////////////////////////////////////////////////////////
//	 点名存贮，返回点名对应的点号
int  CCoGPS_GetStationNumber(char *buf,struct CCoGPS *ccgps)
{
    for(int i=0; i<ccgps->Pnumber; i++) 
	{
		if(ccgps->Pname[i]!=NULL)
		{
			//与已经存入点名地址数组的点名比较
			if(strcmp(buf,ccgps->Pname[i])==0)return i;
		}
		else
		{
			//将新点名存到内存，地址放到Pname数组中
			int len = strlen(buf);
			ccgps->Pname[i] = (char *)malloc(sizeof(char)*(len+1));//new char[len+1];
			strcpy(ccgps->Pname[i], buf);
			return i;
		}
	}
	return -1;
}


//////////////////////////////////////////////////////////////////////////
//   GPS网抗差估计
void CCoGPS_Robust(int fname,double eps,struct CCoGPS *ccgps)
{
	int n=3*ccgps->Vnumber;    //观测值总数
	ccgps->QV=(double *)malloc(sizeof(double)*n);//new double [n];  //残差的权倒数
	ccgps->W =(double *)malloc(sizeof(double)*n);//new double [n];  //权因子
	
	//------------------------------------
	//  计算观测值的权矩阵
	int ii=0;  // 计数器
	for(int i=0; i<ccgps->Anumber; i++)
	{
		int ni = 3*(ccgps->dir0[i+1]-ccgps->dir0[i]); //向量组观测值总数
		for(int j=0;j<ni;j++)
		{
			ccgps->QV[ii]=ccgps->P[i][ij(j,j)];
			ii++;
		}
		inverse(ccgps->P[i],ni);
	}
	//--------------------------------------
	//  计算残差的权倒数
	ccgps->CaATPA(ccgps);   //组成法方程
	if(!inverse(ccgps->ATPA,3*ccgps->Pnumber))  //法方程系数矩阵求逆
	{
		fprintf(stderr,"法方程系数降秩降秩！\n");
		exit(0);
	}
	ccgps->caQV(ccgps);  //计算残差的权倒数
	
	//--------------------------------------
	//  将观测值的权备份到数组PP	
	double **PP= (double **)malloc(sizeof(double *)*ccgps->Anumber);//new double *[ccgps->Anumber]; //等价权数组
	for(int k=0; k<ccgps->Anumber; k++)
	{
		int nk=3*(ccgps->dir0[k+1]-ccgps->dir0[k]);
		PP[k]=(double *)malloc(sizeof(double)*nk*(nk+1)/2);//new double[nk*(nk+1)/2];
		for(int i=0;i<nk*(nk+1)/2;i++) PP[k][i]=ccgps->P[k][i];
	}
	
	//---------------------------------------
	// 抗差估计迭代
	for(int i=0;i<n;i++) ccgps->W[i]=1.0;//权因子赋初值
	int No=0; //迭代次数
	while(1)
	{
		No++;
		// 计算等价权
		for(int k=0; k<ccgps->Anumber; k++) // k－向量组编号
		{
			int nk = 3*(ccgps->dir0[k+1]-ccgps->dir0[k]); //观测值个数
			double *Wk=ccgps->W+3*ccgps->dir0[k];         //权因子的数组地址
			
			int index = 0;
			for(int i=0;i<nk;i++)
			{
				for(int j=0;j<=i;j++)
				{
					if(Wk[i]<1.0 || Wk[j]<1.0)
						ccgps->P[k][index] = PP[k][index]*sqrt(Wk[i]*Wk[j]);
					else ccgps->P[k][index] = PP[k][index];
					index++;
				}
			}
		}

		ccgps->CaATPA(ccgps); //组成法方程
		ccgps->Known(ccgps);  // 已知点处理
		double maxX=ccgps->Ca_dX(0,ccgps);     
		ccgps->CaLV(ccgps->V,ccgps);  //计算残差

		//fprintf(resultfp,"\n抗差估计，第%d次迭代...\nmax(dX)=%e",No,maxX);
		// 		PrintLV("抗差估计残差",V);
		
		if(fabs(maxX)<eps && No>1)
		{
			break;
		}

		//-----------------------------------------------
		//权因子计算
		ccgps->Ca_UintV(ccgps);  //标准化残差计算(放在W里)
		for(int i=0;i<3*ccgps->Vnumber;i++)
		{
			ccgps->W[i]=Wi(fname,ccgps->W[i]/ccgps->MEDIAN,5.0 , 8.0);
		}
	}

	int dd=0; //降权观测值的个数
	for(int i=0;i<3*ccgps->Vnumber;i++)
	{
		if(ccgps->W[i]<1.0)dd++;
	}
	
	if(dd==0)
	{
		fprintf(ccgps->resultfp,"\n  所有观测值均未降权!\n");	
	}
	else
	{   //输出降权因子
		fprintf(ccgps->resultfp,"\n  共有%d个观测值被降权!\n",dd);	
		fprintf(ccgps->resultfp,"\n\n权因子: \n");
		for(int i=0;i<3*ccgps->Vnumber;i++)
		{
			if(ccgps->W[i]<1.0)
			{
				int ni=ccgps->AreaNumber[i/3];
				fprintf(ccgps->resultfp,"\n%8d",ni+1);
				fprintf(ccgps->resultfp,"%10s %10s",
					ccgps->Pname[ccgps->dir1[i/3]],ccgps->Pname[ccgps->dir2[i/3]]);
				fprintf(ccgps->resultfp,"   d%c %10.4lf",'x'+i%3,ccgps->W[i]);
			}
		}  
	}
				
	double VTPV=ccgps->CaVTPV(ccgps);
	ccgps->Sigma=sqrt(VTPV/(n-dd-3*ccgps->unPnumber));
	ccgps->PrintXYZ(ccgps);
	ccgps->PrintLV("V",ccgps->V,ccgps);

	fprintf(ccgps->resultfp,"\n\n抗差估计:\n[pvv]=%e",VTPV);
	fprintf(ccgps->resultfp,"\n单位权中误差μ=±%e",ccgps->Sigma);	
	
	free(PP);
	//PrintM(resultfp,W,3*Vnumber,3,"%6.3lf ","迭代权因子");	
}

//////////////////////////////////////////////////////////////////////////
//     拟稳平差
void CCoGPS_Quasi_Stable(char *file,struct CCoGPS *ccgps)
{
	//将观测值的方差协方差矩阵求逆，化为权矩阵
	for(int i=0; i<ccgps->Anumber; i++)
	{
		int ni = 3*(ccgps->dir0[i+1]-ccgps->dir0[i]); //向量组观测值总数
		inverse(ccgps->P[i],ni);
	}
	
	ccgps->CaATPA(ccgps);
	
	FILE *fp=fopen(file,"r");
	if(fp==NULL) 
	{
		fprintf(stderr,"拟稳点信息文件打不开！\n");
		exit(0);
	}
	
	bool *IsStable = (bool *)malloc(sizeof(bool)*ccgps->Pnumber);//new bool[ccgps->Pnumber]; //是否为拟稳点
	for(int i=0;i<ccgps->Pnumber;i++)
	{
		IsStable[i]=false;
	}
	int StableNumber; //拟稳点总数
	fscanf(fp,"%d",	&StableNumber);
	for(int i=0;i<StableNumber;i++)
	{
		char buf[50];
		fscanf(fp,"%s",buf);
		int k=ccgps->GetStationNumber(buf,ccgps);
		if(k<0)
		{
			fprintf(stderr,"拟稳点名文件中点名错误！\n"); 
			exit(0);
		}
		IsStable[k]=true;
	}
	fclose(fp);

	double f=1.0/StableNumber;
	for(int i=0; i<3*ccgps->Pnumber; i++)
	{
		if(!IsStable[i/3]) continue;//表示拟稳点，跳过去
		for(int j=0; j<=i;j++)
		{
			if(i%3==j%3 && IsStable[j/3])
				ccgps->ATPA[ij(i,j)]+=f;
		}
	}
	
	free(IsStable);
	ccgps->Ca_dX(true,ccgps);
	
	for(int i=0; i<3*ccgps->Pnumber; i++)
	{
		for(int j=0; j<=i;j++)
		{
			if(i%3==j%3) ccgps->ATPA[ij(i,j)]-=f;
		}
	}
	
	ccgps->CaLV(ccgps->V,ccgps);
	double VTPV=ccgps->CaVTPV(ccgps);
	ccgps->Sigma = sqrt(VTPV/(3*ccgps->Vnumber-3*(ccgps->Pnumber-1)));
	
	ccgps->PrintXYZ(ccgps);
	fprintf(ccgps->resultfp,"\n\n拟稳平差:\n[pvv]=%e",VTPV);
	fprintf(ccgps->resultfp,"\n单位权中误差μ=±%e",ccgps->Sigma);	
}




///////////////////////////////////////////////////////////////
// 计算粗差估值的权逆阵Qg,粗差估值gross,返回粗差二次型VTPVg
double CCoGPS_CaQg(
			int    s,      //粗差参数的个数
			int    Jint[], //加了粗差参数的观测值序号（集）
			double gross[],//粗差向量
			double Qg[],   //粗差权逆阵
			double JTPA[],  //PA的子矩阵
			double JTPV[],  //V的子向量
			struct CCoGPS *ccgps)
{
     ccgps->CaJTPJ( Qg, Jint, s,ccgps);
     ccgps->CaJTPA(JTPA,Jint[s-1],s,ccgps);	 

	 // 粗差估值的权逆阵计算
     int t=3*ccgps->Pnumber;   //参数总数
	 for(int j=0;j<s; j++) 
	 { 
		 for(int i=0; i<=j; i++)
		 {
			 for(int k=0; k<t; k++)
			 {
				 double a=JTPA[j*t+k];
				 if(a+1.0==1.0)continue;
				 for(int l=0;l<t;l++)
				 {
					 Qg[ij(j,i)]-=a*ccgps->ATPA[ij(k,l)]*JTPA[i*t+l];
				 }
			 }
		 }
	 }  
	 if(!inverse2(Qg,s))return -2.0;
	 
	 // 计算粗差估值
	 JTPV[s-1]=ccgps->PV[Jint[s-1]];
	 for(int i=0; i<=s-1; i++)
     	{
		 gross[i]=0.0;
		 for(int j=0;j<=s-1;j++)
			 gross[i]-=Qg[ij(i,j)]*JTPV[j];
	 }

     // 计算粗差二次型
	 double VTPVg=0.0;
     for(int i=0; i<=s-1; i++)
	 {
		 VTPVg-=JTPV[i]*gross[i];
	 }
     return VTPVg;
}

#include "probability.h"

//////////////////////////////////////////////////////////////////////////
//    粗差探测函数      2008-08-29
void CCoGPS_DataSnooping(double Arfa,int Max,struct CCoGPS *ccgps)  // Max-最大搜索次数
{
     //  为数组申请内存
     double *JTPA=(double *)malloc(sizeof(double)*3*ccgps->Pnumber*Max);//new double [3*ccgps->Pnumber*Max];
     double *JTPV=(double *)malloc(sizeof(double)*Max);//new double [Max];
     double *Qg= (double *)malloc(sizeof(double)*Max*(Max+1)/2);//new double [Max*(Max+1)/2]; // 粗差权逆阵
     double *gross= (double *)malloc(sizeof(double)*Max);//new double [Max];  // 粗差估值
     int    *Jint= (int *)malloc(sizeof(int)*Max);//new int [Max];  // 可疑观测值序号

     //CProbability MyProb; //概率计算类
	 
	 //-------------------------------------------------
	 //  最小二乘平差
	 for(int i=0; i<ccgps->Anumber; i++)
	 {
		 int ni = 3*(ccgps->dir0[i+1]-ccgps->dir0[i]); //向量组观测值总数
		 inverse(ccgps->P[i],ni);
	 }
	 
	 ccgps->CaATPA(ccgps);    
	 ccgps->Known(ccgps);	 
	 ccgps->Ca_dX(false,ccgps);     
	 ccgps->CaLV(ccgps->V,ccgps);  //计算最小二乘残差
	 double VTPV=ccgps->CaVTPV(ccgps);

	 //---------------------------------------------------
	 //粗差探测
     fprintf(ccgps->resultfp,"\n           逐个搜索法粗差探测:\n");

     int maxp;  //当前搜索中，统计量最大的观测值序号
     int Gn=0;  //当前搜索次数
	 double  new_Omg;      // 本次搜索最大粗差二次型;
     double  last_Omg=0.0; // 前次搜索最大粗差二次型
  	 int r=3*ccgps->Vnumber-3*ccgps->unPnumber; // 多余观测数		 
	 
	 bool *W=(bool *)malloc(sizeof(bool)*3*ccgps->Vnumber);//new bool[3*ccgps->Vnumber];  //粗差标志数组
	 for(int i=0; i<3*ccgps->Vnumber; i++)ccgps->W[i]=0;
			 
     while(1)
     {
        if(Gn+1>Max)
		{
			fprintf(stderr,"给定的搜索次数过小，请增大搜索次数，重新开始搜索。\n");
			fclose(ccgps->resultfp);
			exit(0);
		}

  	    fprintf(ccgps->resultfp,"\n第%d次搜索:",Gn+1);
		
        double Flimit=probability_re_F(1,r-(Gn+1),1.0-Arfa);

        new_Omg=0.0; 
        for(int i=0; i<3*ccgps->Vnumber; i++)
		{
			 Jint[Gn]=i;

             if(ccgps->W[i]) continue;	// 检查号观测值是否为已知粗差观测值

	     double Omg = ccgps->CaQg(Gn+1,Jint,gross,Qg,JTPA,JTPV,ccgps); //粗差二次型
             if(Omg<0.0)continue;

             if(Omg>new_Omg && Omg<VTPV)
			 {
				 new_Omg=Omg; 
				 maxp=i;  
			 }
		}

        double F=(new_Omg-last_Omg)/(VTPV-new_Omg)*(r-(Gn+1));
		   
		if(F>Flimit)fprintf(ccgps->resultfp,"\n   F=%6.3lf(>%6.3lf)",F,Flimit);
		else	    fprintf(ccgps->resultfp,"\n   F=%6.3lf(<%6.3lf)",F,Flimit);

		Jint[Gn]=maxp;

		double Omg=ccgps->CaQg(Gn+1,Jint,gross,Qg,JTPA,JTPV,ccgps);
 
		for(int i=0;i<Gn+1;i++)
		{
			int j=Jint[i];
			fprintf(ccgps->resultfp,"\n   No.%-2d",j);
			fprintf(ccgps->resultfp,"  向量组:%-3d ",ccgps->AreaNumber[j/3]+1); 
			fprintf(ccgps->resultfp,"%5s ",ccgps->Pname[ccgps->dir1[j/3]]);
			fprintf(ccgps->resultfp,"%5s ",ccgps->Pname[ccgps->dir2[j/3]]);
			fprintf(ccgps->resultfp," (%c)",'X'+j%3);
			fprintf(ccgps->resultfp,"%12.4lf",gross[i]);
		}

        if( F<=Flimit || fabs(gross[Gn])<0.01)break;
		Gn++;
        last_Omg=new_Omg;
		ccgps->W[maxp]=1;
    }

    if(Gn==0)
    {
	      fprintf(ccgps->resultfp,"\n\n粗差探测结果：未发现粗差");
    }
	else
	{
		 for(int j=1;j<=Gn;j++)
		 {
			 new_Omg=ccgps->CaQg(j,Jint,gross,Qg,JTPA,JTPV,ccgps);
		 }
		 fprintf(ccgps->resultfp,"\n\n 逐个搜索法最后定位:(粗差改正)\n");
		 fprintf(ccgps->resultfp," 粗差总数: %d\n",Gn);
		 
		 for(int i=0;i<Gn;i++)
		 {
			int j=Jint[i];
			fprintf(ccgps->resultfp,"\n   No.%-2d",j);
			fprintf(ccgps->resultfp,"  向量组:%-3d ",ccgps->AreaNumber[j/3]+1); 
			fprintf(ccgps->resultfp,"%5s ",ccgps->Pname[ccgps->dir1[j/3]]);
			fprintf(ccgps->resultfp,"%5s ",ccgps->Pname[ccgps->dir2[j/3]]);
			fprintf(ccgps->resultfp," (%c)",'X'+j%3);
			fprintf(ccgps->resultfp,"%12.4lf",gross[i]);
		 }
		 
		 fprintf(ccgps->resultfp,"\n\n剔除粗差后最小二乘平差结果");
		 VTPV=VTPV-new_Omg;
		 ccgps->Sigma=sqrt(VTPV/(r-Gn));
		 ccgps->Ca_X_Qx(gross,Qg,JTPA,Gn,ccgps);
		 
		 ccgps->PrintXYZ(ccgps);
		 ccgps->CaLV(ccgps->V,ccgps);//计算残差
		 ccgps->PrintLV("V",ccgps->V,ccgps);	  
		 fprintf(ccgps->resultfp,"\n\n粗差探测:\n[pvv]=%e",VTPV);
		 fprintf(ccgps->resultfp,"\n单位权中误差μ=±%e",ccgps->Sigma);	
		 
	}


	free(JTPA);
	free(JTPV);
	free(Qg);
	free(gross);
	free(Jint);
	free(W);
}


//////////////////////////////////////////////////////////////////////
//        计算JTPA的最后一行(n0-1)
//         JTPA是n0行t列的矩阵
void CCoGPS_CaJTPA(double JTPA[],int ks,int s,struct CCoGPS *ccgps) 
//s-可疑观测值总数 ;  ks-最后一个可疑观测值的编号
{
     int t=3*ccgps->Pnumber;
     for(int i=0;i<t;i++)JTPA[(s-1)*t+i]=0.0;

	 int k =ccgps->AreaNumber[ks/3];  // 第ks号观测值的组号
     int k1=ccgps->dir0[k];           // k组首向量的序号
     int k2=ccgps->dir0[k+1];         // k+1组区首向量的序号
	 int nk=3*(k2-k1);

	 int s2=ks-3*k1; // 组内观测值编号
	 int *Dir1=ccgps->dir1+k1;
	 int *Dir2=ccgps->dir2+k1;
	 
	 for(int s1=0;s1<nk;s1++)
     	{
         int i1=3*Dir1[s1/3]+s1%3; 
         int i2=3*Dir2[s1/3]+s1%3; 
		 
		 double p12 = ccgps->P[k][ij(s1,s2)];
		 		
		 JTPA[(s-1)*t+i1]-=p12;
         JTPA[(s-1)*t+i2]+=p12;
     	}

	 /*   
     int num=AreaNumber[ks/3];  // 第ks号观测值的组号
     int k1=dir0[num];         //  num组首向量的序号
     int k2=dir0[num+1];       //  num+1区首向量的序号


     int j=ks-3*k1;
     for(int k=3*k1;k<3*k2;k++)
     {
         int i=k-3*k1;
         int ij=(j<=i) ? i*(i+1)/2+j : j*(j+1)/2+i;
         double pij = P[num][ij];

         int i1=3*dir1[k/3]+k%3;
         int i2=3*dir2[k/3]+k%3;
         JTPA[(s-1)*t+i1]-=pij;
         JTPA[(s-1)*t+i2]+=pij;
     }

*/

}


//////////////////////////////////////////////////////////////////////
//	计算JTPJ，s-粗差参数的个数
void CCoGPS_CaJTPJ(double JTPJ[], int Jint[],int s,struct CCoGPS *ccgps)
{
	for(int k=0; k<s; k++)
	{
		int k1=Jint[k]; //观测值编号
		int A1=ccgps->AreaNumber[k1/3]; //观测值所属同步区编号
		double *Pk = ccgps->P[A1];      //本同步区权矩阵数组
		int k0 = 3*ccgps->dir0[A1];     //本同步区首观测值的编号    			

		for(int i=0;i<=k;i++)
		{ 
			int k2 = Jint[i]; //观测值编号
			if(ccgps->AreaNumber[k2/3] == A1) //k1,k2属于同一同步区
			{
				JTPJ[ij(k,i)]=Pk[ij(k1-k0,k2-k0)];
			}
			else JTPJ[ij(k,i)]=0.0;
		}	  
	}
}


//////////////////////////////////////////////////////////////////////////
//   计算剔除s个粗差后坐标的平差值及权逆阵
void CCoGPS_Ca_X_Qx(double gross[],double Qg[],double JTPA[],int s,struct CCoGPS *ccgps)
// gross[] -粗差向量
// Qg[] - 粗差估值权逆阵
// J'PA 
// s      粗差个数
{
	int t=3*ccgps->Pnumber;

	// M = inverse(N) * A'PJ
	double *M=(double *)malloc(sizeof(double)*t*s);//new double[t*s];
	for(int i=0;i<t*s;i++) M[i]=0.0;
	for(int i=0;i<t;i++)
	{
		for(int j=0;j<s;j++)
		{
			for(int k=0; k<t; k++)
				M[i*s+j]+=ccgps->ATPA[ij(i,k)]*JTPA[j*t+k];
		}
	}

	for(int i=0;i<t;i++)
	{
		double xi=0.0;
		for(int j=0;j<s;j++)
			xi+=M[i*s+j]*gross[j];
		ccgps->XYZ[i]-=xi;
	}

	for(int i=0;i<t; i++)
	{
		for(int j=0;j<=i; j++)
		{
			double nij=0.0;
			for(int k=0; k<s; k++)
				for(int h=0; h<s; h++)
					nij+=M[i*s+k]*Qg[ij(k,h)]*M[j*s+h];
			ccgps->ATPA[ij(i,j)]+=nij;
		}
	}
	free(M);
	return;
}




//////////////////////////////////////////////////////////////////////////
//   输入已知点的坐标、方差－协方差矩阵(已知点相关)、中误差(已知点独立)
bool CCoGPS_InputKnownXYZ(char *file,bool correlative,struct CCoGPS *ccgps)
{
	ccgps->m_IsPx = true; //参数加权平差

    FILE *fp=fopen(file,"r");  
	if(fp==NULL )
	{
		fprintf(stderr,"已知坐标文件打不开\n");
		return 0;
	}
	
	int kn;
	fscanf(fp,"%d",&kn);   //已知点总数
	ccgps->unPnumber=ccgps->Pnumber-kn;
	
	int tt=(3*kn)*(3*kn+1)/2;   //权矩阵元素的个数
	ccgps->KnownPoints=(int *)malloc(sizeof(int)*kn);//new int [kn];   //已知点点号
	ccgps->KnownXYZ=(double *)malloc(sizeof(double)*3*kn);//new double [3*kn]; //已知点坐标
	ccgps->KnownPx=(double *)malloc(sizeof(double)*tt);//new double [tt];    //已知点坐标的权矩阵
	for(int i=0;i<tt;i++) ccgps->KnownPx[i]=0.0;
	for(int i=0;i<ccgps->Pnumber;i++)ccgps->IsKnown[i]=false; //初始值：未知点

	if(correlative)//已知坐标是相关的
	{
		for(int i=0;i<kn;i++)
		{
			char name[20];
			fscanf(fp,"%s%lf%lf%lf",name,
				ccgps->KnownXYZ+3*i,ccgps->KnownXYZ+3*i+1,ccgps->KnownXYZ+3*i+2); 
			int k=ccgps->GetStationNumber(name,ccgps);
			if(k<0)
			{
				fprintf(stderr,"读取已知坐标文件出错：点名错误！\n");  
				return 0;
			}
			
			ccgps->KnownPoints[i]=k; //点号
			ccgps->IsKnown[k]=true;
		} 
		
		int jj=0;
		for(int i=0;i<3*kn;i++)
		{
			int d;
			fscanf(fp,"%d",&d);   // 读取行号
			for(int j=0;j<=i;j++)
			{ 
				fscanf(fp,"%lf",ccgps->KnownPx+jj); //方差协方差
				jj++;
			}
		} 
		
		fprintf(ccgps->resultfp,"\n 已知点坐标:\n");
		for(int i=0;i<kn;i++)
		{
			int k=ccgps->KnownPoints[i];
			fprintf(ccgps->resultfp,"\n%10s",ccgps->Pname[k]);
			fprintf(ccgps->resultfp," %15.5lf %14.5lf %14.5lf",
				ccgps->KnownXYZ[3*i],ccgps->KnownXYZ[3*i+1],ccgps->KnownXYZ[3*i+2]);
		}
		PrintM2(ccgps->resultfp,ccgps->KnownPx,3*kn,4,"%18.10e ","已知点坐标的方差阵",true);

		bool rn=inverse(ccgps->KnownPx,3*kn); //方差协方差矩阵求逆，的权矩阵
		if(!rn) 
		{
			fprintf(stderr,"输入已知坐标文件出错：坐标方差阵降秩\n");
			return 0;
		}
	}
	else //已知坐标是独立的
	{
		for(int i=0;i<kn;i++)
		{
			char name[20];
			fscanf(fp,"%s%lf%lf%lf",name,
				ccgps->KnownXYZ+3*i,ccgps->KnownXYZ+3*i+1,ccgps->KnownXYZ+3*i+2); 
			
			int k=ccgps->GetStationNumber(name,ccgps);
			if(k<0)
			{
				fprintf(stderr,"读取已知坐标文件出错：点名错误！\n");    
				return 0;
			}
			
			ccgps->KnownPoints[i]=k;
			ccgps->IsKnown[k]=true;
						
			double mx,my,mz;
			fscanf(fp,"%lf%lf%lf",&mx,&my,&mz);
			ccgps->KnownPx[ij(3*i,3*i)] = 1.0/(mx*mx);
			ccgps->KnownPx[ij(3*i+1,3*i+1)] = 1.0/(my*my);
			ccgps->KnownPx[ij(3*i+2,3*i+2)] = 1.0/(mz*mz);
		} 
		fprintf(ccgps->resultfp,"\n 已知点坐标及其中误差:\n");
		for(int i=0;i<kn;i++)
		{
			int k=ccgps->KnownPoints[i];
			fprintf(ccgps->resultfp,"\n%10s",ccgps->Pname[k]);
			fprintf(ccgps->resultfp," %15.5lf %14.5lf %14.5lf",
				ccgps->KnownXYZ[3*i],ccgps->KnownXYZ[3*i+1],ccgps->KnownXYZ[3*i+2]);

			double mx=ccgps->KnownPx[ij(3*i,3*i)];
			double my=ccgps->KnownPx[ij(3*i+1,3*i+1)];
			double mz=ccgps->KnownPx[ij(3*i+2,3*i+2)];
			fprintf(ccgps->resultfp," %6.4lf %6.4lf %6.4lf",
							sqrt(1.0/mx),sqrt(1.0/my),sqrt(1.0/mz));
		}
	}
	fclose(fp);	
	ccgps->unPnumber=ccgps->Pnumber-kn;
	return 1;
}

