// NetworkAdjust.cpp: implementation of the CNetworkAdjustment class.
//
//////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <stdbool.h>
#include "planenetadjust.h"

#include "public.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

void CPlaneNetAdjust_CPlaneNetAdjust(struct CPlaneNetAdjust *clna)
{
	clna->m_Lnumber=0;
	clna->m_Snumber=0;
	clna->m_Tnumber=0;
	clna->m_Pnumber=0;
	clna->CPlaneNetAdjust=CPlaneNetAdjust_CPlaneNetAdjust;
	clna->de_CPlaneNetAdjust=CPlaneNetAdjust_de_CPlaneNetAdjust;
	clna->InputData=CPlaneNetAdjust_InputData;
	clna->PrintData=CPlaneNetAdjust_PrintData;
	clna->GetStationNumber=CPlaneNetAdjust_GetStationNumber;
	clna->Get_S12=CPlaneNetAdjust_Get_S12;
	clna->Get_T12=CPlaneNetAdjust_Get_T12;
	clna->Get_Angle=CPlaneNetAdjust_Get_Angle;
	clna->ca_xy0=CPlaneNetAdjust_ca_xy0;
	clna->ca_x0y0=CPlaneNetAdjust_ca_x0y0;
	clna->ca_T12=CPlaneNetAdjust_ca_T12;
	clna->ca_ab=CPlaneNetAdjust_ca_ab;
	clna->ca_cd=CPlaneNetAdjust_ca_cd;
	clna->ca_ATPA=CPlaneNetAdjust_ca_ATPA;
	clna->ca_ATPAi=CPlaneNetAdjust_ca_ATPAi;
	clna->ca_dX=CPlaneNetAdjust_ca_dX;
	clna->ca_V=CPlaneNetAdjust_ca_V;
	clna->qq=CPlaneNetAdjust_qq;
	clna->PrintResult=CPlaneNetAdjust_PrintResult;
	clna->LS_Adjust=CPlaneNetAdjust_LS_Adjust;
	clna->Free=CPlaneNetAdjust_Free;
	clna->Quasi_Stable=CPlaneNetAdjust_Quasi_Stable;
	clna->ca_GT=CPlaneNetAdjust_ca_GT;
	clna->ErrorEllipse=CPlaneNetAdjust_ErrorEllipse;
	clna->Snooping=CPlaneNetAdjust_Snooping;
}

void CPlaneNetAdjust_de_CPlaneNetAdjust(struct CPlaneNetAdjust *clna)
{
	if(clna->m_Lnumber>0) //方向观测值数组初始化
	{
		free(clna->dir0);
		free(clna->dir1);    
		free(clna->dir2);
		free(clna->L);
		free(clna->V); 
	}
	if(clna->m_Snumber>0)//边长观测值数组初始化
	{
		free(clna->S_dir1);
		free(clna->S_dir2);
		free(clna->S_L);
		free(clna->S_V);
	}
	
	if(clna->m_Tnumber>0)//方位角观测值数组初始化
	{
		free(clna->T_dir1);
		free(clna->T_dir2);
		free(clna->T_L);
		free(clna->T_V);
	}
	if(clna->m_Pnumber>0)//方位角观测值数组初始化
	{
		free(clna->XY);
		for(int i=0;i<clna->m_Pnumber;i++)
		{
			free(clna->Pname[i]);
		}
		free(clna->Pname);
	}		
}



//////////////////////////////////////////////////////////////////////////
//	   点号存贮、根据点名换算点号
int  CPlaneNetAdjust_GetStationNumber(char *name,struct CPlaneNetAdjust *clna)
{
	int i=0;
	for(i=0;i<clna->m_Pnumber;i++)
	{
		if(clna->Pname[i]==NULL) break;
		if(strcmp(name,clna->Pname[i])==0)return i;
	}

	if(i < clna->m_Pnumber )//已经编过点号的点数小于总点数
	{
        clna->Pname[i]=(char *)malloc(sizeof(char)*(strlen(name)+1));//new char[strlen(name)+1];
        strcpy(clna->Pname[i],name);
        return i;
	}
	else 
	{
		fprintf(clna->resultfp,"点名错误，程序中断执行！");
	}
	return -1;
}



//////////////////////////////////////////////////////////////////////////
//   读取原始数据文件
void CPlaneNetAdjust_InputData(char *file,struct CPlaneNetAdjust *clna)
{
	FILE *fp;
    if((fp=fopen(file,"r"))==NULL)
	{    
		fprintf(stderr,"无法打开原始数据文件！\n"); 
		exit(0);
	}

    fscanf(fp,"%d%d%d%d",&clna->m_Pnumber, // 总点数
			      &clna->m_knPnumber,       // 已知点数
			      &clna->m_Lnumber,         // 方向值组数
			      &clna->m_Nnumber          // 方向值总数
			      );   

    fscanf(fp,"%d%d",&clna->m_Snumber,  // 边长点数
			      &clna->m_Tnumber      // 方位角数
				  );   
 	
	clna->XY=(double *)malloc(sizeof(double)*2*clna->m_Pnumber);//new double   [2*clna->m_Pnumber]; 
	
    	clna->Pname=(char **)malloc(sizeof(char *)*clna->m_Pnumber);//new char* [clna->m_Pnumber];
	for(int i=0;i<clna->m_Pnumber;i++)clna->Pname[i]=NULL;
	 
	if(clna->m_Lnumber>0) // 为方向观测值数组申请内存
	{
		 clna->dir0=(int *)malloc(sizeof(int)*(clna->m_Lnumber+1));  //new int [clna->m_Lnumber+1]; //各组首方向观测值的序号
		 clna->dir1=(int *)malloc(sizeof(int)*(clna->m_Lnumber));  //new int [clna->m_Lnumber];   //测站点号
		 clna->dir2=(int *)malloc(sizeof(int)*(clna->m_Lnumber));  //new int [clna->m_Nnumber];   //照准点号
		 clna->L=(double*)malloc(sizeof(double)*(clna->m_Lnumber));  //new double [clna->m_Nnumber];   //方向值
		 clna->V=(double *)malloc(sizeof(double)*(clna->m_Lnumber));  //new double [clna->m_Nnumber];   //粗差，V之前，放自由项l
	}
	if(clna->m_Snumber>0)//为边长观测值数组申请内存
	{
		 clna->S_dir1=(int *)malloc(sizeof(int)*(clna->m_Snumber));  //new int[clna->m_Snumber];  //测站点号
		 clna->S_dir2=(int *)malloc(sizeof(int)*(clna->m_Snumber));  //new int[clna->m_Snumber];  //照准点号
		 clna->S_L=(double *)malloc(sizeof(double)*(clna->m_Snumber));  //new double[clna->m_Snumber];  //边长观测值
		 clna->S_V=(double *)malloc(sizeof(double)*(clna->m_Snumber));  //new double[clna->m_Snumber];  //边长残差
	}

	if(clna->m_Tnumber>0)//为方位角观测值数组申请内存
	{
		 clna->T_dir1=(int *)malloc(sizeof(int)*(clna->m_Snumber));//new int[clna->m_Snumber];  //测站点号
		 clna->T_dir2=(int *)malloc(sizeof(int)*(clna->m_Snumber));//new int[clna->m_Snumber];  //照准点号
		 clna->T_L=(double *)malloc(sizeof(double)*(clna->m_Snumber));//new double[clna->m_Snumber];  //方位角观测值
		 clna->T_V=(double *)malloc(sizeof(double)*(clna->m_Snumber));//new double[clna->m_Snumber];  //方位角残差
	}

	// 观测值可用标志数组
	int n=clna->m_Nnumber+clna->m_Snumber+clna->m_Tnumber;
	clna->Usable=(bool *)malloc(sizeof(bool)*n);//new bool[n];
	for(int i=0;i<n; i++)clna->Usable[i]=true;

    int t=2*clna->m_Pnumber+clna->m_Lnumber; // 未知参数总数
    int tt=t*(t+1)/2;

    clna->ATPL=(double *)malloc(sizeof(double)*t);//new double [t];   //法方程自由项
    clna->ATPA=(double *)malloc(sizeof(double)*tt);//new double [tt];  //系数矩阵
    clna->dX=(double *)malloc(sizeof(double)*t);//new double [t];     //未知数向量
	 
    int unPnumber=clna->m_Pnumber-clna->m_knPnumber;
	 
	// 读取观测值中误差
	fscanf(fp,"%lf%lf%lf%lf",&clna->ma,&clna->mS1,&clna->mS2,&clna->mT);
	
	char name[100];
	for(int i=0;i<clna->m_knPnumber;i++)//读取已知点坐标
    	{
	     	fscanf(fp,"%s",name);
         	int k=clna->GetStationNumber(name,clna);
	     	fscanf(fp,"%lf%lf",clna->XY+2*k, clna->XY+2*k+1);
   	 }
	
	if(clna->m_Lnumber>0) //读取方向值
	{
		clna->dir0[0]=0;
		for(int i=0;i<=clna->m_Lnumber-1;i++) 
		{
			int ni; // ni: 测站方向数
			fscanf(fp,"%s%d",name,&ni); 
			
			if(ni<1) 
			{
				fprintf(stderr,"\n数据文件错误：方向数小于1!\n"); 
				exit(0);
			}
			 
			clna->dir1[i]=clna->GetStationNumber(name,clna);
			clna->dir0[i+1]=clna->dir0[i]+ni;
			for(int j=clna->dir0[i];j<clna->dir0[i+1];j++)
			{
				fscanf(fp,"%s%lf",name,clna->L+j);
				clna->dir2[j]=clna->GetStationNumber(name,clna); //照准点号
				clna->L[j]=dms_rad(clna->L[j]); // dms_arc: 度分秒化弧度
				
			}
		} 
	}

    for(int i=0;i<=clna->m_Snumber-1;i++)  //读取边长
	{
		fscanf(fp,"%s",name);   
		clna->S_dir1[i]=clna->GetStationNumber(name,clna);

		fscanf(fp,"%s",name);   
		clna->S_dir2[i]=clna->GetStationNumber(name,clna);

		fscanf(fp,"%lf",&clna->S_L[i]);
	} 
	 
    for(int i=0;i<=clna->m_Tnumber-1;i++)  //读取方位角
    {
		fscanf(fp,"%s",name);   
		clna->T_dir1[i]=clna->GetStationNumber(name,clna);

		fscanf(fp,"%s",name);   
		clna->T_dir2[i]=clna->GetStationNumber(name,clna);

		fscanf(fp,"%lf",&(clna->T_L[i]));
		clna->T_L[i]=dms_rad(clna->T_L[i]);
    }

    fclose(fp);
}

//////////////////////////////////////////////////////////////////////////
//  原始数据输出
void CPlaneNetAdjust_PrintData(struct CPlaneNetAdjust *clna)
{
     fprintf(clna->resultfp," 总点数: %d\n 方向组总数: %d\n 方向值总数: %d\n",
					 clna->m_Pnumber,clna->m_Lnumber,clna->m_Nnumber);
     fprintf(clna->resultfp," 边长总数: %d\n 方位角总数: %d  \n",
						clna->m_Snumber,clna->m_Tnumber);
	 
     fprintf(clna->resultfp,"\n   已知点坐标:\n");
     for(int i=0;i<clna->m_knPnumber;i++)
     {
	     fprintf(clna->resultfp,"\n%8s  ",clna->Pname[i]);
	     fprintf(clna->resultfp,"%12.3lf%14.3lf",clna->XY[2*i],clna->XY[2*i+1]);
     }

     if(clna->m_Lnumber>0)
	 {
		 fprintf(clna->resultfp,"\n\n   方向观测值:\n");
		 for(int i=0;i<=clna->m_Lnumber-1;i++)
		 {
     		  int k1=clna->dir1[i];
			  for(int j=clna->dir0[i];j<clna->dir0[i+1];j++)
			  {
				if(j==clna->dir0[i])
				{
					fprintf(clna->resultfp,"%10s  %d \n",
						clna->Pname[k1],clna->dir0[i+1]-clna->dir0[i]);
				}
				fprintf(clna->resultfp,"%15s %15.3lf\n",
							clna->Pname[clna->dir2[j]],rad_dms(clna->L[j]));
			  }
		 }
	 }

	 if(clna->m_Snumber>0)
	 {
		 fprintf(clna->resultfp,"\n   边长观测值:\n");
		 for(int i=0;i<clna->m_Snumber;i++)
		 {
			 int k1=clna->S_dir1[i];
			 int k2=clna->S_dir2[i];
			 fprintf(clna->resultfp,"%10s %10s  %11.4lf\n",
				            clna->Pname[k1],clna->Pname[k2],clna->S_L[i]);
		 }
	 }
	 
	 if(clna->m_Tnumber>0)
	 {
		 fprintf(clna->resultfp,"\n\n 方位角观测值:\n");
		 for(int i=0;i<=clna->m_Tnumber-1;i++)
		 {
			 int k1=clna->T_dir1[i];
			 int k2=clna->T_dir2[i];
			 fprintf(clna->resultfp,"%10s %10s  %15.3lf\n",
					 clna->Pname[k1],clna->Pname[k2],rad_dms(clna->T_L[i]));
		 }
	 } 
	 
	 fflush(clna->resultfp);
}


//////////////////////////////////////////////////////////////////////////
//   根据三点号返回以p为顶点的一个夹角
double CPlaneNetAdjust_Get_Angle(int p,int p1,int p2,struct CPlaneNetAdjust *clna)
{
	double PAI=M_PI;
	
      int i,j,k,i1,i2;
      double A;

      for(i=0;i<=clna->m_Lnumber-1;i++)
      {
	     i1=i2=-1;
	     k=clna->dir1[i];
	     for(j=clna->dir0[i];j<clna->dir0[i+1];j++)
	     {
		      if(k!=p)break;
		      if(p1==clna->dir2[j])i1=j;
		      if(p2==clna->dir2[j])i2=j;
	     }
	     if(i1>=0&&i2>=0)
	     {
		      A=clna->L[i2]-clna->L[i1];
		      if(A<0.0)A+=2.0*PAI;
		      return A;
	     }
      }
      return -1.0;    //找不到符合条件的角，返回负值
}

//////////////////////////////////////////////////////////////////////////
//        计算近似坐标(三角网)
void CPlaneNetAdjust_ca_xy0(struct CPlaneNetAdjust *clna)
{
     int i,j,jj,k,k1,k2,s,f;
     double A,B,C,xA,yA,xB,yB;
     double sinA,sinB,cosA,cosB,sinC,xk,yk;
	 int unPnumber=clna->m_Pnumber-clna->m_knPnumber;

     for(i=0;i<clna->m_knPnumber;i++)clna->XY[2*i]=1.0e30;  //设置坐标未知点标志

      s=0;
      while(1)
      {
        s++;
	     for(i=0;i<=clna->m_Lnumber-1;i++)
	     {
	       k=clna->dir1[i];
	       if(clna->XY[2*k]<1.0e20)continue;
	       for(j=clna->dir0[i];j<clna->dir0[i+1];j++)
	       {
	          if(clna->XY[2*k]<1.0e20)break;
	          k2=clna->dir2[j];
	          xB=clna->XY[2*k2];
	          yB=clna->XY[2*k2+1];
	          if(xB>1.0e29)continue;

	          for(jj=j+1;jj<clna->dir0[i+1];jj++)
	          {
		           k1=clna->dir2[jj];
		           xA=clna->XY[2*k1];
		           yA=clna->XY[2*k1+1];
		           if(xA>1.0e29)continue;

			   A=clna->Get_Angle(k1,k,k2,clna);
		           B=clna->Get_Angle(k2,k1,k,clna);
		           C=clna->L[jj]-clna->L[j];
		           if(A<0.0||B<0.0)continue;

			   double PAI=3.14159265358979312;
				   
		           sinA=sin(A);		  sinB=sin(B);		  sinC=sin(C);
         		   if(C>PAI)
		           {
		              cosB=cos(B);
		              xk=xB+((xA-xB)*cosB-(yA-yB)*sinB)*sinA/sinC;
		              yk=yB+((xA-xB)*sinB+(yA-yB)*cosB)*sinA/sinC;
		           }
		           else
		           {
		                cosA=cos(A);
		                xk=xA+((xB-xA)*cosA+(yB-yA)*sinA)*sinB/sinC;
		                yk=yA+((yB-yA)*cosA-(xB-xA)*sinA)*sinB/sinC;
		           }
					  
				   clna->XY[2*k]=xk;
				   clna->XY[2*k+1]=yk;
	               break;
	          } //jj
	       } //j
	     }
	     f=1;
	     for(i=0;i<clna->m_Pnumber;i++)if(clna->XY[2*i]>1.0e29){ f=0; break; }
	     if(f)break;
	     if(s>unPnumber && !f)
	     {
	        printf("\n有下列点无法计算出坐标:\n");
	        for(i=0;i<clna->m_Pnumber;i++)
		            if(clna->XY[2*i]>1.0e29)printf("\n%s",clna->Pname[i]);
	        exit(0);
		//getch();///////////////////////////////////////////////////////////////////////////??????
	     }
      }

	   fprintf(clna->resultfp,"\n近似坐标\n");
		
	   for(i=0;i<=clna->m_Pnumber-1;i++)
	   { 
			 double xi=clna->XY[2*i];
			 double yi=clna->XY[2*i+1];
			 fprintf(clna->resultfp,"\n%2d %3s ",i+1,clna->Pname[i]);
			 fprintf(clna->resultfp,"%14.3lf%12.3lf", xi,yi+500000.0);
	   } 
	   fflush(clna->resultfp);		
}


//////////////////////////////////////////////////////////////////////////
//        计算方向误差方程式ab系数
double CPlaneNetAdjust_ca_ab(int k1,int k2,double A[],int Ain[],struct CPlaneNetAdjust *clna)
{
	const double ROU=2.062648062470963630e+05;///////////////////////////////////////////////////?????
	const double PAI=M_PI;
	
	
	double dx=clna->XY[2*k2]-clna->XY[2*k1];
	double dy=clna->XY[2*k2+1]-clna->XY[2*k1+1];
	
    double s2=dx*dx+dy*dy;

    A[0]= dy/s2*ROU;     Ain[0]=2*k1;
    A[1]=-dx/s2*ROU;     Ain[1]=2*k1+1;
    A[2]=-dy/s2*ROU;     Ain[2]=2*k2;
    A[3]= dx/s2*ROU;     Ain[3]=2*k2+1;

    double T=atan2(dy,dx);
    if(T<0.0)T=T+2.0*PAI;
    return T;
}

//////////////////////////////////////////////////////////////////////////
//       计算边长误差方程系数
double CPlaneNetAdjust_ca_cd(int k1,int k2,double A[],int Ain[],struct CPlaneNetAdjust *clna)
{
	
	double dx=clna->XY[2*k2]-clna->XY[2*k1];
	double dy=clna->XY[2*k2+1]-clna->XY[2*k1+1];
    	double s = sqrt(dx*dx+dy*dy);
	
    	A[0]=-dx/s;     Ain[0]=2*k1;/////////////////////////////////////////////////////////////////////////////?????
    	A[1]=-dy/s;     Ain[1]=2*k1+1;
    	A[2]= dx/s;     Ain[2]=2*k2;
    	A[3]= dy/s;     Ain[3]=2*k2+1;
    	return s;
}

//////////////////////////////////////////////////////////////////////////
//	一个误差方程式组成的法方程
void CPlaneNetAdjust_ca_ATPAi(double B[],int Bin[],double p,double Li,int m,struct CPlaneNetAdjust *clna)
{
    int k,s,i,j;
    double ai,aj;

    for(k=0;k<m;k++)
    {
	    i=Bin[k];
	    ai=B[k];
	    clna->ATPL[i]-=p*ai*Li;////////////////////////////////////////////////////////////////////////////////?????可能有溢出
	    for(s=0;s<m;s++)
	    {
		j=Bin[s];
	        if(i>j)continue;
	        aj=B[s];
	        clna->ATPA[ij(i,j)]+=p*ai*aj;//////////////////////////////////////////////////////////////////////?????可能有溢出，此处溢出
		} 
	} 
} 

//////////////////////////////////////////////////////////////////////////
//         组 成 法 方 程 式
void CPlaneNetAdjust_ca_ATPA(struct CPlaneNetAdjust *clna)
{
	const double ROU=2.062648062470963630e+05;
	const double PAI=M_PI;	
	
	double B[5];
    	int Bin[5];
	
	int t=2*clna->m_Pnumber+clna->m_Lnumber;
    	int tt=t*(t+1)/2;
    	for(int i=0;i<=tt-1;i++) clna->ATPA[i]=0.0;
    	for(int i=0;i<=t-1;i++) clna->ATPL[i]=0.0;

	//---------------------------------------
	//  方向值组成法方程
	double Pi=1.0/(clna->ma*clna->ma);
	B[4]=-1.0;
    	for(int i=0;i<=clna->m_Lnumber-1;i++)
	{
		int k1=clna->dir1[i];
		Bin[4]=2*clna->m_Pnumber+i; //定向角改正数的未知数编号 /////////////////////////////////////////////////////////?????
		
		double z;     //定向角近似值
		for(int j=clna->dir0[i];j<clna->dir0[i+1];j++)
		{
			if(!clna->Usable[j])continue;
			int k2=clna->dir2[j];
			double T=clna->ca_T12(k1,k2,clna); //返回值：方位角
			z=T-clna->L[j];
			if(z<0.0) z+=2.0*PAI;
			break;
		}
		
		for(int j=clna->dir0[i];j<clna->dir0[i+1];j++)
		{
			int k2=clna->dir2[j];
			double T12=clna->ca_ab(k1,k2,B,Bin,clna);/////////////////////////////////////////////////////////////???????
			double Lj=T12-clna->L[j];
			if(Lj<0.0)Lj+=2.0*PAI;
			Lj=(Lj-z)*ROU;
			clna->V[j]=Lj; //自由项放在V数组里，计算残差时使用
			if(clna->Usable[j]) 
			{
				clna->ca_ATPAi(B,Bin,Pi,Lj,5,clna);////////////////////////////////////////////j=10时clna->S_dir1[0]>14000
			}
		}    
	} 
	
	//---------------------------------------
	//  边长组成法方程
	for(int i=0;i<=clna->m_Snumber-1;i++)
	{
		if(!clna->Usable[i+clna->m_Nnumber])continue;
		int k1=clna->S_dir1[i];
		int k2=clna->S_dir2[i];

		double m=clna->mS1+clna->mS2*clna->S_L[i];
		double Pi=1.0/(m*m+1.0e-15); // 边长的权
		  
		double S12=clna->ca_cd(k1,k2,B,Bin,clna);//////////////////////////////////////////////////////////////////?????段错误
		double Li=S12-clna->S_L[i];  
		clna->ca_ATPAi(B,Bin,Pi,Li,4,clna);
	} 
	
	//---------------------------------------
	//  方位角组成法方程
	Pi=1.0/(clna->mT*clna->mT+1.0e-15); //方位角的权
    	for(int i=0;i<=clna->m_Tnumber-1;i++)
	{
		if(!clna->Usable[i+clna->m_Nnumber+clna->m_Snumber])continue;
		int k1=clna->T_dir1[i];
		int k2=clna->T_dir2[i];
		  
		double T12=clna->ca_ab(k1,k2,B,Bin,clna);
		double Li=(T12-clna->T_L[i])*ROU;  
		clna->ca_ATPAi(B,Bin,Pi,Li,4,clna);
	} 
}


//////////////////////////////////////////////////////////////////////////
//  参数平差值计算,返回值：坐标改正数的最大值
double CPlaneNetAdjust_ca_dX(struct CPlaneNetAdjust *clna)
{
     int t=2*clna->m_Pnumber+clna->m_Lnumber; //未知数个数
     if(!inverse(clna->ATPA,t)) //法方程系数矩阵求逆
	 {
		 fprintf(stderr,"调用ca_dX函数出错：法方程系数阵不满秩！\n");
		 fclose(clna->resultfp);
		 exit(0);
	 }

     double max=0.0; //坐标改正数的最大值
     for(int i=0;i<t;i++)
     {
		  double xi=0.0;
		  for(int j=0;j<t;j++)
		  {
			  xi+=clna->ATPA[ij(i,j)]*clna->ATPL[j];
		  }
		  clna->dX[i]=xi;
		  if(i<2*clna->m_Pnumber)
		  {
			  clna->XY[i]+=xi;
			  if(fabs(xi)>max)max=fabs(xi);
		  }
     }
	 return max;
}

//////////////////////////////////////////////////////////////////////////
//     计算最小二乘残差及[pvv]
double  CPlaneNetAdjust_ca_V(struct CPlaneNetAdjust *clna)
{
	clna->m_pvv=0.0;

	double A[5];
	int Ain[5];

	//---------------------------------------
	//  方向值残差计算
	A[4]=-1.0;

	double Pi=1.0/(clna->ma*clna->ma); //方向值的权
	for(int i=0;i<=clna->m_Lnumber-1;i++)
	{
		int k1=clna->dir1[i];
		Ain[4]=2*clna->m_Pnumber+i;

		for(int j=clna->dir0[i];j<clna->dir0[i+1];j++)
		{
			int k2=clna->dir2[j];
			
			double T=clna->ca_ab(k1,k2,A,Ain,clna);

			double vj=clna->V[j];
			for(int s=0;s<5;s++)
			{
				int k=Ain[s]; 
				vj+=A[s]*clna->dX[k];
			}
			clna->V[j]=vj;
			if(clna->Usable[j])
				clna->m_pvv += vj*vj*Pi;
		}
	}

	//---------------------------------------
	//  边长残差计算
	for(int i=0;i<=clna->m_Snumber-1;i++)
	{
		int k1=clna->S_dir1[i];
		int k2=clna->S_dir2[i];

		double S12=clna->ca_cd(k1,k2,A,Ain,clna);

		double vi=S12-clna->S_L[i];
		clna->S_V[i] = vi;
		double m=clna->mS1+clna->mS2*clna->S_L[i];
		Pi=1.0/(m*m);

		if(clna->Usable[clna->m_Nnumber+i])
			clna->m_pvv += Pi*vi*vi;
	}
	  
	//---------------------------------------
	//  方位角残差计算
	Pi=1.0/(clna->mT*clna->mT);
	for(int i=0;i<=clna->m_Tnumber-1;i++)
	{
		int k1=clna->T_dir1[i];
		int k2=clna->T_dir2[i];

		double T12=clna->ca_T12(k1,k2,clna);

		double vi=(T12-clna->T_L[i])*206264.806247;  ///////////////////////////////////////////////////////?????可恶的硬编码
		clna->T_V[i] = vi;
		if(clna->Usable[clna->m_Nnumber+clna->m_Snumber+i])
			clna->m_pvv += Pi*vi*vi;
	}
	return clna->m_pvv;
}


#include "probability.h"

//////////////////////////////////////////////////////////////////////////
//     粗差探测计算函数
double  CPlaneNetAdjust_Snooping(double arfa,struct CPlaneNetAdjust *clna)
{
	//caxy0();
	clna->ca_x0y0(clna);
	//CProbability prty;
	double U=probability_re_norm(1.0-arfa/2.0);
	double snm_pvv;//此处可能有逻辑错误
	int no=0; //粗差搜索次数
	while(1)
	{
		clna->ca_ATPA(clna);
		for(int i=0;i<clna->m_knPnumber;i++) 
		{
			clna->ATPA[ij(2*i,2*i)]+=1.0e20;
			clna->ATPA[ij(2*i+1,2*i+1)]+=1.0e20;
		}
		double max=clna->ca_dX(clna);
		snm_pvv=clna->ca_V(clna);////////////////////////////////////////////////////////??????以m开头，很容易混淆的,和类里的重名了！！！m_pvv
		
		double F[5];
		int Fin[5];
		
		double max_v=0.0;
		int    max_i;

		F[4]=-1.0;
		double m2 = clna->ma*clna->ma;  //方向观测值的中误差平方
		for(int i=0;i<=clna->m_Lnumber-1;i++)
		{
			int k1=clna->dir1[i];
			Fin[4]=2*clna->m_Pnumber+i;
			
			for(int j=clna->dir0[i];j<clna->dir0[i+1];j++)
			{
				if(!clna->Usable[j])continue;
				int k2=clna->dir2[j];
				
				clna->ca_ab(k1,k2,F,Fin,clna);
				
				double q=Calculate_q_compressed(clna->ATPA,F,Fin,5);
				double mv = sqrt(m2-q);
				double vj=clna->V[j]/(mv+1.0e-12);
				if(fabs(vj)>max_v)
				{
					max_v=fabs(vj);
					max_i=j;
				}
			}
		}
		
		for(int i=0;i<=clna->m_Snumber-1;i++)
		{
			if(!clna->Usable[i+clna->m_Nnumber])continue;
			int k1=clna->S_dir1[i];
			int k2=clna->S_dir2[i];
			
			double S12=clna->ca_cd(k1,k2,F,Fin,clna);
			
			double m=clna->mS1+clna->mS2*clna->S_L[i];
			
			double q=Calculate_q_compressed(clna->ATPA,F,Fin,4);
			double mv = sqrt(m*m-q);
			double v=clna->S_V[i]/(mv+1.0e-12);
			if(fabs(v)>max_v)
			{
				max_v=fabs(v);
				max_i=clna->m_Nnumber+i;
			}
		}
		
		m2 = clna->mT*clna->mT;
		for(int i=0;i<=clna->m_Tnumber-1;i++)
		{
			if(!clna->Usable[i+clna->m_Nnumber+clna->m_Snumber])continue;
			int k1=clna->T_dir1[i];
			int k2=clna->T_dir2[i];
			
			double T12=clna->ca_T12(k1,k2,clna);
			
			double q=Calculate_q_compressed(clna->ATPA,F,Fin,4);
			double mv = sqrt(m2-q);
			double v=clna->T_V[i]/(mv+1.0e-12);
			if(fabs(v)>max_v)
			{
				max_v=fabs(v);
				max_i=clna->m_Nnumber+clna->m_Snumber+i;
			}
		}

		if(max_v>U)
		{
			clna->Usable[max_i]=false;
			no++;  //粗差个数
			continue;
		}

		int n=clna->m_Nnumber+clna->m_Snumber+clna->m_Tnumber-no;
		int t=2*(clna->m_Pnumber-clna->m_knPnumber)+clna->m_Lnumber;
		clna->m_mu=sqrt(snm_pvv/(n-no-t));/////////////////////////////////////////////////////////////////?????可恶的m_pvv我将其改为了snm_pvv

		fprintf(clna->resultfp,"\n  粗差探测: μ=±%.3lf",clna->m_mu);
		clna->PrintResult(clna);
		clna->ErrorEllipse(clna);

		if(no>0)fprintf(clna->resultfp,"\n粗差个数：%d",no);
		else    fprintf(clna->resultfp,"\n无粗差");

		break;
	}
		
	return snm_pvv;
}




//////////////////////////////////////////////////////////////////////////
//   平差成果输出
void  CPlaneNetAdjust_PrintResult(struct CPlaneNetAdjust *clna)
{
	int i,j,k1,k2;
	double m1,xi,yi,dxi,dyi,m2;

	fprintf(clna->resultfp,"\n\n            ====  坐标平差值及其精度 ====\n");
	fprintf(clna->resultfp,"\nNo.   P        X            Y      "
		"   mx     my       M\n");

	for(i=0;i<=clna->m_Pnumber-1;i++)
	{
		xi=clna->XY[2*i];
		yi=clna->XY[2*i+1];
		dxi=clna->dX[2*i];
		dyi=clna->dX[2*i+1];
		fprintf(clna->resultfp,"\n%2d %3s ",i+1,clna->Pname[i]);
		fprintf(clna->resultfp,"%14.3lf %12.3lf", xi,yi);

		m1=sqrt(clna->ATPA[ij(2*i,2*i)])*clna->m_mu;
		fprintf(clna->resultfp,"%7.3lf",m1);
		m2=sqrt(clna->ATPA[ij(2*i+1,2*i+1)])*clna->m_mu;
		fprintf(clna->resultfp," %7.3lf",m2);
		fprintf(clna->resultfp," %7.3lf",sqrt(m1*m1+m2*m2));
	}

	if(clna->m_Lnumber>0)
	{
		fprintf(clna->resultfp,"\n\n%20s","");
		fprintf(clna->resultfp,"====   方向观测值平差成果  ====\n");
		fprintf(clna->resultfp,"\n    P1     P2        L        V"
					"        T         mT        S       ms");
		for(i=0;i<=clna->m_Lnumber-1;i++)
		{
			k1=clna->dir1[i];
			for(j=clna->dir0[i];j<clna->dir0[i+1];j++)
			{
				double T,S,q,A[5];
				int Ain[5];
				k2=clna->dir2[j];
				if(j==clna->dir0[i])fprintf(clna->resultfp,"\n\n%6s",clna->Pname[k1]);
				else fprintf(clna->resultfp,"\n      ");
				fprintf(clna->resultfp,"%7s",clna->Pname[k2]);
				fprintf(clna->resultfp,"%13.2lf%7.2lf",rad_dms(clna->L[j]),clna->V[j]);

				T=clna->ca_ab(k1,k2,A,Ain,clna);
				q=sqrt(clna->qq(A,Ain,clna))*clna->m_mu;
				fprintf(clna->resultfp,"%13.2lf %5.2lf",rad_dms(T),q);

				S=clna->ca_cd(k1,k2,A,Ain,clna);
				q=sqrt(clna->qq(A,Ain,clna))*clna->m_mu;
				fprintf(clna->resultfp,"%12.3lf %6.3lf",S,q);
				if(!clna->Usable[j])fprintf(clna->resultfp," 有粗差");
			}
		} 
	} 

	if(clna->m_Snumber>0)
	{
		fprintf(clna->resultfp,"\n\n%25s"," ");
		fprintf(clna->resultfp,"====   边长观测值平差成果  ====\n");
		fprintf(clna->resultfp,"\n    P1     P2         S        Vs"
		"          T         mT        S+Vs     ms");

		double A[5];
		int Ain[5];

		for(i=0;i<=clna->m_Snumber-1;i++)
		{
			int k1=clna->S_dir1[i];
			int k2=clna->S_dir2[i];

			fprintf(clna->resultfp,"\n%6s",clna->Pname[k1]);
			fprintf(clna->resultfp,"%7s",clna->Pname[k2]);
			fprintf(clna->resultfp,"%13.4lf %8.4lf ",clna->S_L[i],clna->S_V[i]);


			double T=clna->ca_ab(k1,k2,A,Ain,clna);
			double prmT=sqrt(clna->qq(A,Ain,clna))*clna->m_mu;///////////////////////////////////////////???????可恶的mT,改为prmT
			fprintf(clna->resultfp,"%13.3lf %6.3lf",rad_dms(T),prmT);///////////////////////////////////////////???????可恶的mT,改为prmT
			double S=clna->ca_cd(k1,k2,A,Ain,clna);
			double prmS=sqrt(clna->qq(A,Ain,clna))*clna->m_mu;///////////////////////////////////////////???????可恶的mS,改为prmS
			fprintf(clna->resultfp,"%12.3lf  %6.3lf",S,prmS);///////////////////////////////////////////???????可恶的mS,改为prmS

			if(!clna->Usable[clna->m_Nnumber+i])fprintf(clna->resultfp," 有粗差");
		} 
	}

	if(clna->m_Tnumber>0)
	{
		fprintf(clna->resultfp,"\n\n%20s","");
		fprintf(clna->resultfp,"====   方位角观测值平差成果  ====\n");
		fprintf(clna->resultfp,"\n    P1     P2        T          V"
		"         T+V         mT        S        ms");

		double A[5];
		int Ain[5];

		for(i=0;i<=clna->m_Tnumber-1;i++)
		{
			int k1=clna->T_dir1[i];
			int k2=clna->T_dir2[i];

			fprintf(clna->resultfp,"\n%6s",clna->Pname[k1]);
			fprintf(clna->resultfp,"%7s",clna->Pname[k2]);
			fprintf(clna->resultfp,"%14.3lf %8.4lf ",	rad_dms(clna->T_L[i]),clna->T_V[i]);

			double T=clna->ca_ab(k1,k2,A,Ain,clna);
			double prmT=sqrt(clna->qq(A,Ain,clna))*clna->m_mu;///////////////////////////////////////////???????可恶的mT,改为prmT
			fprintf(clna->resultfp,"%13.3lf %6.3lf",rad_dms(T),prmT);///////////////////////////////////////////???????可恶的mT,改为prmT
			double S=clna->ca_cd(k1,k2,A,Ain,clna);
			double prmS=sqrt(clna->qq(A,Ain,clna))*clna->m_mu;///////////////////////////////////////////???????可恶的mS,改为prmS
			fprintf(clna->resultfp,"%12.3lf  %6.3lf",S,prmS);///////////////////////////////////////////???????可恶的mS,改为prmS
			if(!clna->Usable[clna->m_Nnumber+clna->m_Snumber+i])
				fprintf(clna->resultfp," 有粗差");
		}
	} 
}



//////////////////////////////////////////////////////////////////////////
//            平面方位角计算
double CPlaneNetAdjust_ca_T12(int k1,int k2,struct CPlaneNetAdjust *clna)  // 返回值为弧度值,
{
	//const double PAI=M_PI;

	double dx=clna->XY[2*k2]-clna->XY[2*k1];
	double dy=clna->XY[2*k2+1]-clna->XY[2*k1+1];

	double T=atan2(dy,dx);
	if(T<0.0)T=T+2.0*M_PI;

	return T;
}



//////////////////////////////////////////////////////////////////////////
//            权 倒 数 计 算
double CPlaneNetAdjust_qq(double B[],int Bin[],struct CPlaneNetAdjust *clna)
{
      int i,j,k1,k2;
      double q=0.0;

      for(i=0;i<4;i++)
      {
	      k1=Bin[i];
	      for(j=0;j<4;j++)
		  {
		       k2=Bin[j];
		       q+=clna->ATPA[ij(k1,k2)]*B[i]*B[j];
		  }
      }

	  if(fabs(q)<1.0e-8) q=0.0; // 因为收舍误差q接近0时可能为负数
      return q;
}



//////////////////////////////////////////////////////////////////////////
//       边角网拟稳平差
int CPlaneNetAdjust_Quasi_Stable(char *file,struct CPlaneNetAdjust *clna)
{
	clna->ca_ATPA(clna); //组成法方程

	double *GT = clna->ca_GT(file,clna); 
	
	int kn=4;
	if(clna->m_Snumber>0)kn--;
	if(clna->m_Tnumber>0)kn--;
	
	int t=2*clna->m_Pnumber+clna->m_Lnumber;
	int *GT_in=(int *)malloc(sizeof(int)*t);//new int[t];
	for(int i=0;i<t;i++)GT_in[i]=i;

	double *GT2=(double *)malloc(sizeof(double)*t*4);//new double[4*t];
	for(int i=0;i<4*t;i++)GT2[i]=0.0;
	
	for(int i=0;i<kn; i++)
	{
		for(int j=0;j<clna->m_Pnumber;j++)
		{
			if(clna->IsStable[j])
			{
				GT2[i*t+2*j]=GT[i*t+2*j];
				GT2[i*t+2*j+1]=GT[i*t+2*j+1];
			}
		}
	}
	
	for(int k=0; k<kn; k++)
	{
		clna->ca_ATPAi(GT2+k*t,GT_in,1.0,0.0,t,clna);
	}
	
	clna->ca_dX(clna);

	for(int k=0; k<kn; k++)
	{
		clna->ca_ATPAi(GT+k*t,GT_in,-1.0,0.0,t,clna);
	}

	clna->m_pvv=clna->ca_V(clna);
	int n=clna->m_Nnumber+clna->m_Snumber+clna->m_Tnumber;
	clna->m_mu=sqrt(clna->m_pvv/(n-2.0*(clna->m_Pnumber-2)-clna->m_Lnumber));

	fprintf(clna->resultfp,"\n\n  水平网拟稳平差: μ=±%lf",clna->m_mu);
	clna->PrintResult(clna);
	
	PrintM2(clna->resultfp, clna->ATPA, 2*clna->m_Pnumber,5,"%13.4e ","坐标平差值之权逆阵",true);
	
	return 1;
}	




//////////////////////////////////////////////////////////////////////////
//   计算G矩阵(转置)
double * CPlaneNetAdjust_ca_GT(char *file,struct CPlaneNetAdjust *clna)
{
	clna->IsStable = (bool *)malloc(sizeof(bool)*clna->m_Pnumber);//new bool[clna->m_Pnumber]; // 拟稳点标志
	
	if(file==NULL) // 自由网平差，不需要拟稳点名文件
	{
		clna->m_StableNumber=clna->m_Pnumber;  //拟稳点总数等于总点数
		for(int i=0;i<clna->m_Pnumber;i++)
		{
			clna->IsStable[i]=true;
		}
	}
	else
	{
		FILE *fp=fopen(file,"r");
		if(fp==NULL)
		{
			fprintf(stderr,"拟稳点名文件打不开:%s\n",file);
			fclose(clna->resultfp);
			exit(0);
		}
		
		for(int i=0;i<clna->m_Pnumber;i++)
		{
			clna->IsStable[i]=false;
		}
		
		fscanf(fp,"%d",	&clna->m_StableNumber);//拟稳点总数
		if(clna->m_StableNumber<2)
		{
			fprintf(stderr," 拟稳点总数小于2！\n");
			fclose(clna->resultfp);
			exit(0);
		}
		
		for(int i=0;i<clna->m_StableNumber;i++)
		{
			char buf[50];
			fscanf(fp,"%s",buf);
			int k=clna->GetStationNumber(buf,clna);
			if(k<0)
			{
				fprintf(stderr,"拟稳点名表文件中存在错误点名!\n");
				fclose(clna->resultfp);
				exit(0);			
			}
			clna->IsStable[k]=true;
		}
		fclose(fp);
	}

	int t=clna->m_Lnumber+2*clna->m_Pnumber;

	double *GT= (double *)malloc(sizeof(double)*t*4);//new double[4*t];//G矩阵的转置
	for(int i=0; i<4*t; i++) GT[i]=0.0;

	//--------------------------------------
	//   计算坐标的均值
	double xm=0.0;
	double ym=0.0;
	for(int i=0;i<clna->m_Pnumber;i++)
	{
		if(clna->IsStable[i])
		{
			xm+=clna->XY[2*i];
			ym+=clna->XY[2*i+1];
		}
	}
	xm/=clna->m_StableNumber;
	ym/=clna->m_StableNumber;
	
	double H=0.0;
	for(int i=0;i<clna->m_Pnumber;i++)
	{
		if(!clna->IsStable[i])continue;
		
		double dx=clna->XY[2*i]-xm;
		double dy=clna->XY[2*i+1]-ym;
		H+=dx*dx+dy*dy;
	}
	H=sqrt(H);	
	double m=1.0/sqrt(clna->m_StableNumber+0.0);
	 
	// ---------------------------------------------
	//  填充GT
	for( int k=0; k<clna->m_Pnumber;k++)
	{
		GT[2*k]=m;      //第一行
		GT[t+2*k+1]=m;  //第二行
		
		double dx=clna->XY[2*k]-xm;
		double dy=clna->XY[2*k+1]-ym;
		
		//第3行
		GT[2*t+2*k]=-dy/H;
		GT[2*t+2*k+1]=dx/H;
		
		//第4行
		GT[3*t+2*k]  =dx/H;
		GT[3*t+2*k+1]=dy/H;
	}

	// 第3行 定向角对应项
	for(int k=2*clna->m_Pnumber; k<t; k++)
	{
		GT[2*t+k]=206264.806247096/H;
	}

	int kn=4;
	if(clna->m_Snumber>0) kn=3; //有边长观测值时,矩阵的最后一行不要了
	if(clna->m_Tnumber>0) //有方位角观测值时,矩阵的第三行不要了
	{
		kn--;
		for(int i=0;i<t;i++)
			GT[2*t+i]=GT[3*t+i];
	}

	fprintf(clna->resultfp,"\n    G矩阵（转置）: ");
	for(int k=0; k<kn; k++)
	{
		fprintf(clna->resultfp,"\n%2d: ",k);
		for(int i=0;i<t;i++)
		{
			fprintf(clna->resultfp," %7.3lf",GT[k*t+i]);
		}
	}	

/*
	// --------------------------------------
	// 验证法方程系数矩阵与G的乘积是否为零
	double max_b=0.0; 
	for( k=0; k<kn; k++)
	{
		for(int i=0;i<t;i++)
		{
			double b=0.0;
			for(int j=0;j<t;j++)
				b+=ATPA[ij(i,j)]*GT[k*t+j]; 
			if(fabs(b)>max_b)max_b=fabs(b);
		}
	}	
	fprintf(resultfp,"\n    (N×G')=0  ：闭合差 = %e",max_b);
*/

	return GT;
}


//////////////////////////////////////////////////////////////////////////
//    自由网平差计算函数
void CPlaneNetAdjust_Free(struct CPlaneNetAdjust *clna)
{
	clna->ca_ATPA(clna);

	double *GT = clna->ca_GT(NULL,clna); //NULL: 自由网平差的标志
	
	int kn=4;
	if(clna->m_Snumber>0)kn--;
	if(clna->m_Tnumber>0)kn--;

	int t=2*clna->m_Pnumber+clna->m_Lnumber;
	int *GT_in=(int *)malloc(sizeof(int)*t);//new int[t];

	for(int i=0;i<t;i++)GT_in[i]=i;
		
	for(int k=0; k<kn; k++)
	{
		clna->ca_ATPAi(GT+k*t,GT_in,1.0,0.0,t,clna);
	}

	clna->ca_dX(clna);

	for(int k=0;k<kn;k++)
	{
		clna->ca_ATPAi(GT+k*t,GT_in,-1.0,0.0,t,clna);
	}
	free(GT);
	free(GT_in);
	
	clna->m_pvv=clna->ca_V(clna);
	
	int n=clna->m_Nnumber+clna->m_Snumber+clna->m_Tnumber;
	clna->m_mu=sqrt(clna->m_pvv/(n-2.0*(clna->m_Pnumber-2)-clna->m_Lnumber));
	
	fprintf(clna->resultfp,"\n  自由网平差: μ=±%lf",clna->m_mu);
	
	PrintM2(clna->resultfp, clna->ATPA, 2*clna->m_Pnumber,5,"%13.4e ","坐标平差值之权逆阵",true);
	clna->PrintResult(clna);

}



//////////////////////////////////////////////////////////////////////////
//    最小二乘平差
void CPlaneNetAdjust_LS_Adjust(char *data_file,struct CPlaneNetAdjust *clna)
{	
	clna->InputData(data_file,clna);  // 输入原始数据
	
	clna->PrintData(clna);           // 输出原始数据
	//caxy0();
	clna->ca_x0y0(clna);   //近似坐标计算
	
	double max=1.0;
	while(max>0.01)
	{
		
		clna->ca_ATPA(clna);
		for(int i=0;i<clna->m_knPnumber;i++) 
		{
			clna->ATPA[ij(2*i,2*i)]+=1.0e20;
			clna->ATPA[ij(2*i+1,2*i+1)]+=1.0e20;
		}
		max=clna->ca_dX(clna);
	}
	
	clna->m_pvv=clna->ca_V(clna);
	
	int n=clna->m_Nnumber+clna->m_Snumber+clna->m_Tnumber;
	int t=2*(clna->m_Pnumber-clna->m_knPnumber)+clna->m_Lnumber;
	clna->m_mu=sqrt(clna->m_pvv/(n-t));
	
	fprintf(clna->resultfp,"\n最小二乘平差:μ=±%.3lf\n",clna->m_mu);
	clna->PrintResult(clna);
	clna->ErrorEllipse(clna);
		
}



//////////////////////////////////////////////////////////////////////////
//    导线网近似坐标计算函数
void CPlaneNetAdjust_ca_x0y0(struct CPlaneNetAdjust *clna)
{
	int unknow=clna->m_Pnumber-clna->m_knPnumber; //未知点数

	//设置未知点标志,未知点的点号从m_knPnumber开始
    for(int i=clna->m_knPnumber;i<clna->m_Pnumber;i++) clna->XY[2*i]=1.0e30;
	
	for(int No=1; ; No++)
	{
		if(unknow==0) return ;
		 
        if(No>(clna->m_Pnumber-clna->m_knPnumber))
		{
			 fprintf(clna->resultfp,"\n\n部分点计算不出近似坐标:");
			 for(int k=0;k<clna->m_Pnumber;k++)
			 {
				 if(clna->XY[2*k]>1.0e29)
				 {
					 fprintf(clna->resultfp,"\n%s",clna->Pname[k]);
				 }
			 }
			 fclose(clna->resultfp);

			 fprintf(stderr,"部分点计算不出近似坐标！\n");
			 exit(0);
		} 

	    for(int i=0;i<=clna->m_Lnumber-1;i++) // 按方向组循环，遍历方向观测值
		{ 
			int k1=clna->dir1[i];  //测站点号

			double x1=clna->XY[2*k1];   //测站点的x坐标
			double y1=clna->XY[2*k1+1]; //测站点的y坐标
			if(x1>1.0e29) continue; //测站点是未知点，转下一方向组

			int j0=clna->dir0[i]; //本方向组首方向观测值的序号
			for(int j=j0; j<clna->dir0[i+1];j++)
			{
				int k2=clna->dir2[j]; //照准点号
			    double T12;
			    if(clna->XY[2*k1]<1.0e29 && clna->XY[2*k2]<1.0e29) //k1、k2都是已知点
				{ 
			 		T12=clna->ca_T12(k1,k2,clna); //用坐标计算方位角
				}
			    else T12=clna->Get_T12(k1,k2,clna);  //在方位角数组中查找起始方位角
			    if(T12>1.0e29) continue;  //无起始方位角

			    for(int k=j0;k<clna->dir0[i+1];k++)
				{ 
					int k3=clna->dir2[k];
					double x3=clna->XY[2*k3];
					if(x3<1.0e29)continue; //k3是已知点

					double S13=clna->Get_S12(k1,k3,clna); //在边长数组中查找边长
					if( S13 >1.0e29 ) continue; //无边长观测值
					
					double T13=T12+clna->L[k]-clna->L[j];
					x3=S13*cos(T13);
					double y3=S13*sin(T13);

					clna->XY[2*k3]  =x1+x3;
					clna->XY[2*k3+1]=y1+y3;
					unknow--;
				}  
			} 
		} 
	} 
}


//////////////////////////////////////////////////////////////////////////
//    查找边长观测值
double CPlaneNetAdjust_Get_S12(int k1,int k2,struct CPlaneNetAdjust *clna)
{	
	for(int i=0; i<clna->m_Snumber; i++)
	{
		if(k1==clna->S_dir1[i] && k2==clna->S_dir2[i])
		{
			return clna->S_L[i];
		}
		if(k1==clna->S_dir2[i] && k2==clna->S_dir1[i])
		{
			return clna->S_L[i];
		}
	}
	
	return 1.0e30;	
}

//////////////////////////////////////////////////////////////////////////
//      在方位角观测值中查找
double CPlaneNetAdjust_Get_T12(int k1,int k2,struct CPlaneNetAdjust *clna)
{
	const double PAI=M_PI;
	
	for(int i=0; i<clna->m_Tnumber; i++)
	{
		if(k1 == clna->T_dir1[i] && k2 == clna->T_dir2[i])
		{
			return clna->T_L[i];
		}
		if(k1==clna->T_dir2[i] && k2==clna->T_dir1[i])
		{
			double T=clna->T_L[i]+PAI;
			if(T>2.0*PAI)T=T-2.0*PAI;
			return T;
		}
	}

	return 1.0e30;
}



//////////////////////////////////////////////////////////////////////////
//         计算误差椭圆
//   2007-03-20, 为程序教材编的函数.
void CPlaneNetAdjust_ErrorEllipse(struct CPlaneNetAdjust *clna )
{
	const double PAI=M_PI;
	
	double m2=clna->m_mu*clna->m_mu;
	fprintf(clna->resultfp,"\n\n            ==== 点位误差椭圆 ====\n\n");
	fprintf(clna->resultfp,"   点名   椭圆长半轴  椭圆短半轴  长轴方位角\n");
	for(int i=0;i<clna->m_Pnumber; i++)
	{
		double mx2=clna->ATPA[ij(2*i,2*i)]*m2;  //x坐标中误差的平方
		double my2=clna->ATPA[ij(2*i+1,2*i+1)]*m2; //y坐标中误差的平方
		double mxy=clna->ATPA[ij(2*i,2*i+1)]*m2; //xy坐标的协方差

		if(sqrt(mx2+my2)<0.0001) continue;

		double K=sqrt((mx2-my2)*(mx2-my2)+4.0*mxy*mxy);
		
		double E=sqrt(0.5*(mx2+my2+K));  //长轴
		double F=sqrt(0.5*(mx2+my2-K));  //短轴
		
		double A; //误差椭圆长轴的方位角
		if(fabs(mxy)<1.0e-14)  // mxy = 0
		{
			if(mx2>my2) A=0.0;
			else A=0.5*PAI;
		}
		else   // mxy ≠ 0
		{
			A=atan( (E*E-mx2)/mxy);
		}
		if(A<0.0)A+=PAI;

		fprintf(clna->resultfp,"\n %6s  %8.3lf  %10.3lf %13.2lf",
							clna->Pname[i],E,F,rad_dms(A));
	}
}

